# $Id: recorddef.py,v 1.29 2013/05/01 16:32:12 irees Exp $
import collections

from emen2.web.view import View



@View.register
class RecordDef(View):

    @View.add_matcher(r'^/recorddef/(?P<name>[^/]*)/new/$')    
    def new(self, name=None, **kwargs):
        if self.request_method == 'post':
            mainview = kwargs.pop('mainview', '')
            views = {}
            view_name = kwargs.pop('view_name', [])
            view_view = kwargs.pop('view_view', [])
            for k,v in zip(view_name, view_view):
                if k and v:
                    views[k] = v
            kwargs['views'] = views
            if kwargs.get('private') == None: kwargs['private'] = False
            recorddef = self.db.recorddef.new(mainview=mainview, name=name)
            recorddef.update(kwargs)
            rd = self.db.recorddef.put(recorddef)
            if rd:
                self.redirect(self.routing.reverse('RecordDef/main', name=rd.name))
                return
                
        self.main(name=name)
        self.template = '/pages/recorddef'
        self.recorddef.parents = set([self.recorddef.name])
        self.recorddef.children = set()
        self.ctxt['edit'] = True
        self.ctxt['new'] = True
        self.title = 'New Protocol based on: %s'%self.recorddef.desc_short




    @View.add_matcher(r'^/recorddef/(?P<name>[^/]*)/edit/$')    
    def edit(self, name=None, **kwargs):
        if self.request_method == 'post':
            views = {}    
            view_name = kwargs.pop('view_name', [])
            view_view = kwargs.pop('view_view', [])
            for k,v in zip(view_name, view_view):
                if k and v:
                    views[k] = v
            kwargs['views'] = views
            if kwargs.get('private') == None: kwargs['private'] = False
            recorddef = self.db.recorddef.get(name)
            recorddef.update(kwargs)
            rd = self.db.recorddef.put(recorddef)
            if rd:
                self.redirect(self.routing.reverse('RecordDef/main', name=rd.name))
                return
                
        self.main(name=name)
        self.template = '/pages/recorddef'
        self.ctxt['edit'] = True
        self.title = 'Edit Protocol: %s'%self.recorddef.desc_short


    @View.add_matcher(r'^/recorddef/(?P<name>[^/]*)/$')    
    def main(self, name=None):
        self.recorddef = self.db.recorddef.get(name, filt=False)
        self.template = '/pages/recorddef'
        self.title = "Protocol: %s"%self.recorddef.desc_short

        parentmap = self.routing.execute('Tree/embed', db=self.db, keytype='recorddef', root=self.recorddef.name, mode='parents', recurse=3)

        self.ctxt.update(dict(
            parentmap = parentmap,
            editable = self.recorddef.writable(),
            create = self.db.auth.check.create(),
            recorddef = self.recorddef,
            edit = False,
            new = False,
        ))



@View.register
class RecordDefs(View):

    @View.add_matcher(r'^/recorddefs/name/$')
    def name(self, *args, **kwargs):
        return self.main(action='name', *args, **kwargs)


    @View.add_matcher(r'^/recorddefs/count/$')
    def count(self, *args, **kwargs):
        return self.main(action='count', *args, **kwargs)


    @View.add_matcher(r'^/recorddefs/tree/$')
    def tree(self, action=None, q=None, *args, **kwargs):
        return self.main(action='tree', *args, **kwargs)


    @View.add_matcher(r'^/recorddefs/$')
    def main(self, action=None, q=None):
        recorddefnames = self.db.recorddef.filter(None)
        
        if action == None or action not in ["tree", "name", "count"]:
            action = "tree"

        if q:
            recorddefs = self.db.recorddef.find(q)
            action = "name"
        else:
            recorddefs = self.db.recorddef.get(recorddefnames)

        # Tab Switcher
        pages = collections.OrderedDict()
        pages['tree'] = 'Protocol ontology'
        pages['name'] = 'Protocols by name'
        pages['count'] = 'Protocols by number of records'
        uris = {}
        for k in pages:
            uris[k] = self.routing.reverse('RecordDefs/%s'%k)
        pages.uris = uris
        pages.active = action
        self.ctxt['pages'] = pages

        self.template = '/pages/recorddefs.%s'%action        
        self.title = pages.get(action)

        # Children
        childmap = self.routing.execute('Tree/embed', db=self.db, mode="children", keytype="recorddef", root="root", recurse=-1, id='sitemap')

        # Record counts
        count = {}
        if action != 'tree':
            for pd in recorddefs:
                count[pd.name] = len(self.db.record.findbyrectype(pd.name))

        self.ctxt['recorddefnames'] = recorddefnames
        self.ctxt['q'] = q
        self.ctxt['count'] = count
        self.ctxt["recorddefs"] = recorddefs
        self.ctxt["childmap"] = childmap
        self.ctxt['create'] = self.db.auth.check.create()




__version__ = "$Revision: 1.29 $".split(":")[1][:-1].strip()



# $Id: debug.py,v 1.37 2012/07/29 02:30:34 irees Exp $
"""
Classes:
    Output: represents an individual output stream which logs certain specified states
        - Headless: skips the header information and just logs the message
        - Min: represents one which logs everything greater than a specified state
        - stderr: logs to sys.stderr
        - Bounded: represents one which loges everything between two specified states
            - stdout: logs to sys.stdout

    DebugState (NOTE: this probably should be renamed): contains a bunch of logging and debugging utilities.

    debugDict: used by DebugState.instrument_class to log all attribute access of a class

    Filter: subclass of file, used to be use to strip headers off of access.log, unused
"""

from __future__ import with_statement

import functools
import code
import inspect
import os
import os.path
import time
import datetime
import sys
import traceback
from twisted.python import log


__all__ = ['DebugState', 'Output', 'Bounded', 'Headless', 'Min']


def take(num, iter_):
    for _ in range(num): yield iter_.next()

class DebugState(object):
    """Handles logging etc.."""
    def msg(self, msg):
        twisted.python.log.msg(msg, level='DEBUG')

    def start_timer(self, key=''):
        self.msg('timer %s started' % key)
        self._timers[key] = time.time()

    def stop_timer(self, key=''):
        stime = time.time()
        stime = (stime-self._timers.get(key, 0)) * 1000
        self.msg('timer %s stopped: %s milliseconds' % (key, stime))
        return stime

    def note_var(self, var):
        """log the value of a variable and return the variable"""
        self.msg('NOTED VAR:::', var)
        return var

    def loud_notify(self, msg):
        'Notifies and forces attention'
        self.msg(msg)
        raw_input('hit enter to continue...')

    def interact(self, locals, *args):
        """\
        start a interactive console in a given scope
        """
        interp = code.InteractiveConsole(locals)
        interp.interact('Debugging')


    def debug_func(self, func):
        """\
        decorator to print info about a function whenever it is called
        """
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            self.msg('debugging state -> self._state: %r, self._log_state: %r' % (self._state, self._log_state) )
            self.msg('debugging callable: %s, args: %s, kwargs: %s'  % (func, args, kwargs))
            self.print_traceback()
            result = None
            try:
                result = func(*args, **kwargs)
            except BaseException, e:
                self.msg('the callable %(func)r failed with: (%(exc)r) - %(exc)s' % dict(func=func, exc=e))
                raise
            else:
                self.msg('the callable %r returned: %r\n---' % (func, result))
            finally:
                self.pop_state()
                self.last_debugged = (func, args, kwargs)
            return result
        _inner.func = func
        return (_inner if self.debugstates.DEBUG  in (self._state, self._log_state) else func)

    def instrument_class(self, name, bases, dict):
        for nm,meth in [ (nm,meth) for nm,meth in dict.items() if hasattr(meth, '__call__')]:
            self.msg('Instrumenting class (%s) method (%s)' % (name, nm))
            dict[nm] = self.debug_func(meth)
        clss = type(name, bases, dict)
        return clss

    def print_exception(self):
        traceback.print_exc(self)
        return self

    def print_traceback(self, steps=3):
        msg =  self._get_last_module(steps)
        self.msg(msg, tb=False)
        return msg

    def _get_last_module(self, num=1):
        modname = '%s.py' % __name__.split('.')[-1]
        try:
            result = take(num, (
                x for x in (
                    (name[1].split(os.path.sep)[-1], name[0].f_lineno) for name in inspect.getouterframes(inspect.currentframe())
                ) if not x[0]=='debug.py')
            )
        except:
            result = [('Unknown Output',-1)]
        result = [ [str(y) for y in x] for x in result]
        if num == 1:
            result = result[0]
        return result

    def indent(self, string):
        buf = string.split('\n')
        out = [buf.pop(0)]
        for line in buf:
            out.append('\t\t%s' % line)
        return '\n'.join(out)

    def print_list(self, lis, join=' '):
        #deal with Python's unicode brokenness :-{
        def get_right(mess):
            if type(mess) == unicode: pass
            elif type(mess) == str:
                mess = mess.decode('utf-8', 'replace')
            else:
                mess = get_right(str(mess))
            return mess
        return join.join(self.indent(get_right(i)) for i in lis)



class debugDict(dict):
    def __getitem__(self, name):
        result = dict.get(self, name)
        print name, result
        if not self.has_key(name):
            dict.__getitem__(self, name)
        return result
    def get(self, name, default=None):
        result = dict.get(self, name, default)
        print name, result
        return result


#################################################################
########################## OLD STUFF !!! ########################
#################################################################

################
# class Output #
################

class Output(object):
    """Controls access to a paricular output stream

    Subclasses that will be used should be registered with register and created with factory
    """

    subclasses = {}
    @classmethod
    def register_subclass(cls, subcls):
        """A decorator for registering a subclass, uses subcls.__name__.lower() as a key in a lookup table

        :param class subcls: the class to be registered
        :returns: the class passed in as subclass"""

        if issubclass(subcls, cls):
            cls.subclasses[subcls.__name__.lower()] = subcls
        return subcls

    @classmethod
    def factory(cls, version=None, **kwargs):
        """Create an particular kind of instance of Output

        :param str version: the lookup key for the subclass
        :param kwargs: the arguments for the constructor of the desired class
        """

        cls = cls.subclasses.get(version.lower(), cls)
        return cls(**kwargs)


    def __init__(self, states=0, filename=None, file=None, modulename=None, **kwargs):
        """Constructor: don't override this method, override _init"""
        self._states = states
        self._file = file or open(filename, 'a')
        self._state_checked = False
        self._modulename = modulename
        self._init(**kwargs)

    def _not_if_closed(func):
        """decorator: turns methods on or off based on whether the underlying stream is closed"""
        @functools.wraps(func)
        def _inner(self, *args, **kwargs):
            result = None
            if not self.closed:
                result = func(self, *args, **kwargs)
            return result
        return _inner

    @_not_if_closed
    def send(self, module, state, header, msg):
        header, msg = self._preprocess(state, header, msg)

        if self._modulename is not None and self._modulename != module:
            pass

        elif self._state_checked or self.checkstate(state):
            self._file.write(header)
            self._file.write(msg.rstrip())
            self._file.write('\n')
            self.flush()
            self._state_checked = False

    @_not_if_closed
    def flush(self):
        """flush()

        Flush the buffer"""
        if not self._file.closed:
            self._file.flush()

    @_not_if_closed
    def close(self):
        """close()

        Close the buffer"""
        if not self._file.closed:
            self._file.close()

    # subclass behavior modification hooks
    @property
    def closed(self):
        """Indicates whether the underlying stream is open or closed"""
        return self._file.closed

    def checkstate(self, state):
        """Determine whether current message should be logged

        :param state: The current state of the Logger

        This method should set self._state_checked to True if the message is to be logged"""
        result = False
        if self._states is None: result = False
        elif self._states == DebugState.debugstates.ALL: result = True
        elif isinstance(self._states, (str, unicode)): result = False
        else: result = state in self._states
        self._state_checked = result
        return result

    def disable(self):
        """called to disable the stream"""
        self._states = None

    def _init(self, **kwargs):
        """Initialize the instance, override this instead of __init__"""
        pass

    def _preprocess(self, state, header, msg):
        """Called before the message is logged, allows for the displayed message to be changed as one likes


        :param state: the current logging state
        :param str header: the header of the current message
        :param str msg: the text of the current message
        :returns: a tuple (header, msg) containing the desired values of each"""
        return header, msg

@Output.register_subclass
class Min(Output):
    """prints any messages in a state higher than the one given"""
    def checkstate(self, state):
        if self._states is None: result = False
        else:
            result = self._states == DebugState.debugstates.ALL or state >= self._states
            self._state_checked = result
        return result


class Bounded(Output):
    """prints any messages in a given range of states"""
    def _init(self, max, **kwargs):
        self._max = max
    def checkstate(self, state):
        if self._states is None: result = False
        else:
            result = self._states == DebugState.debugstates.ALL or (state >= self._states and state < self._max)
            self._state_checked = result
        return result

@Output.register_subclass
class Headless(Output):
    """drops the header from messages"""
    def _preprocess(self, state, header, msg):
        return '', msg

def stdouts_handler(prefix,suffix):
    def _preprocess(self, state, header, msg):
        return '%s%s'%(prefix, header), '%s%s\n' % (msg.rstrip(),suffix)
        # header = header.split(' : ')
        # head = [header[0]]
        # head.append(header[1]+'\t\t')
        # head.extend(header[2:])
        # header = ' : '.join(head)
        # return state, ' '.join([prefix, header]), msg
    return _preprocess

@Output.register_subclass
class stdout(Bounded):
    _preprocess = stdouts_handler('  ','')

@Output.register_subclass
class stderr(Min):
    _preprocess = stdouts_handler('!!', '') # not sure why you need pre and postpend



class Filter(file):
    def __init__(self, *args, **kwargs):
        self._splitter = kwargs.pop('sep', '::')
        self._num = kwargs.pop('items', 1)
        file.__init__(self, *args, **kwargs)

    def write(self, str_):
        str_ = str_.split(self._splitter, self._num)[-1]
        file.write(self, str_)


#def _get_last_module(debugstate):
#    try:
#        result = filter(lambda x:not x[0]=='debug.py',
#                ( (name[1].split(os.path.sep)[-1], name[0].f_lineno)
#                        for name in inspect.getouterframes(inspect.currentframe())
#                ))[0]
#    except:
#        result = ('Unknown Output',-1)
#    return ':'.join(str(x) for x in result)


__version__ = "$Revision: 1.37 $".split(":")[1][:-1].strip()



# $Id: reverse.py,v 1.6 2012/07/28 06:31:18 irees Exp $
import collections
import jsonrpc.jsonutil

from emen2.web.view import View

# for python >= 2.6
@View.register
class ReverseURI(View):
    # for python < 2.6 do this:

    __matcher__ = dict(
        main=r'^/url/(?P<name>\w+)/$',
        execute=r'^/url/(?P<name>\w+)/(?P<exe>execute)/$',
        alt=r'^/url/$'
    )

    def main(self, arg='', name=None, arguments='', kwargs='', exe=False):
        self.make_raw()
        self.set_header('content-type', 'application/json')

        if name is None and isinstance(arg, (str, unicode)):
            arg = jsonrpc.jsonutil.decode(arg)
        elif name is not None:
            if isinstance(arguments, (str, unicode)):
                arguments = jsonrpc.jsonutil.decode(arguments)
            if isinstance(kwargs, (str, unicode)):
                kwargs = jsonrpc.jsonutil.decode(kwargs)
            arg = [name, arguments, kwargs]

        if exe:
            self.execute_url(name, arguments, kwargs)
        else: self.get_url(arg)

    def get_url(self, arg):
        ret_list = True
        if arg and isinstance(arg[0], (str, unicode) ): arg, ret_list = [arg], False

        arg = [ (name, args, dict( (k.encode('utf-8'),v) for k,v in kwargs.iteritems() ))
                    for name, args, kwargs in arg]

        arg = [(name, self.ctxt.reverse(name, *args, **kwargs)) for name, args, kwargs in arg]

        if ret_list is False:
            if arg: arg = arg[0][1]
            else: arg = ''
        else:
            tmp = collections.defaultdict(list)
            for key, value in arg: tmp[key].append(value)
            arg = tmp

        self.page = jsonrpc.jsonutil.encode(arg)

    def execute_url(self, name, arguments, kwargs):
        kwargs = dict( (k.encode('utf-8'), v) for k,v in kwargs.iteritems() )
        self.page = self.ctxt.render_template_view(name, *arguments, **kwargs)
        
__version__ = "$Revision: 1.6 $".split(":")[1][:-1].strip()



# $Id: __init__.py,v 1.35 2013/05/14 15:52:32 irees Exp $
"""EMEN2: An extesible electronic lab notebook and database."""

def opendb(**kwargs):
    """Open a database."""
    # Import the config first and parse
    import emen2.db.config
    cmd = emen2.db.config.UsageParser()
    import emen2.db.database
    return emen2.db.database.opendb(**kwargs)


def opendbwithopts(optclass, **kwargs):
    import emen2.db.config
    cmd = emen2.db.config.UsageParser(optclass=optclass)
    import emen2.db.database
    return cmd, emen2.db.database.opendb(**kwargs)
        

__version__ = "$Revision: 1.35 $".split(":")[1][:-1].strip()


# $Id: user.py,v 1.29 2013/05/14 14:12:37 irees Exp $
import re
import os
import time
import random
import hashlib

import emen2.db.exceptions
import emen2.db.database
import emen2.util.listops
from emen2.web.view import View


# Use randomly assigned usernames?
HASH_USERNAME_FORCE = False
HASH_USERNAME = True


@View.register
class User(View):

    @View.add_matcher("^/user/(?P<name>[^/]*)/$")
    def main(self, name=None):
        self.template = "/pages/user.view"
        user = self.db.user.get(name, filt=False)
        self.ctxt["user"] = user
        self.title = "User: %s"%(user.displayname)
        if user.disabled:
            self.notify("This user account is disabled", error=True)

    @View.add_matcher("^/user/(?P<name>[^/]*)/edit/$")
    def edit(self, name=None, **kwargs):
        self.template = "/pages/user.edit"
        user = self.db.user.get(name, filt=False)
        self.ctxt["user"] = user
        self.title = "Edit user: %s"%(user.displayname)

        # Security is of course is checked by the database, 
        # this just hides the form itself.
        if self.db.auth.check.context()[0] != user.name and not self.ctxt['ADMIN']:
            raise emen2.db.exceptions.SecurityError, "You may only edit your own user page"

        msgs = {
            "password":"Password updated.", 
            "email":"Email updated.", 
            "privacy":"Privacy updated.", 
            "profile":"Profile information updated."}
            
        msg = msgs.get(kwargs.get('saved'))
        if msg:
            self.notify(msg)

        if self.request_method != 'post':
            return

        u = kwargs.get('user', {})
        if u:
            user.update(u)
            self.db.user.put(user)
            self.simple("Saved account settings.")



@View.register
class Users(View):

    @View.add_matcher(r'^/users/$')    
    def main(self, q=None):
        self.template = "/pages/users"
        self.title = "User directory"
        usernames = self.db.user.filter(None)

        if q:
            users = self.db.user.find(q)
        else:
            users = self.db.user.get(usernames)
        
        if not users:
            self.simple(content='No users found, or insufficient permissions to view user roster.')
            return

        self.ctxt['usernames'] = usernames
        self.ctxt['users'] = users
        self.ctxt['q'] = q or ''

    @View.add_matcher(r'^/users/admin/$', name='admin')    
    def admin(self, sortby="name_last", reverse=0, q=None, **kwargs):
        self.template="/pages/users.admin"
        self.title = 'User administration'
        
        if q:
            users = self.db.user.find(q)
        else:
            users = self.db.user.get(self.db.user.filter(None))

        self.ctxt['args'] = kwargs
        self.ctxt['q'] = q
        self.ctxt['users'] = users
        self.ctxt['sortby'] = sortby
        self.ctxt['reverse'] = reverse



@View.register
class NewUser(View):

    @View.add_matcher("^/users/new/$", view='Users', name='new')
    def new(self, user=None, userrec=None, **kwargs):
        self.template = '/pages/users.new'
        self.title = 'Account request'
        self.ctxt['kwargs'] = kwargs
        self.ctxt['invalid'] = set()

        # Account settings
        user = user or {}
        # Profile settings
        userrec = userrec or {}

        self.ctxt['user'] = user
        self.ctxt['userrec'] = userrec
        
        # Process form if posted.
        if self.request_method != 'post':
            return
                
        # Always required new user parameters.
        REQUIRED = set(['name_last', 'name_first'])

        # Note: other parameters can be marked as required at the FORM level,
        # using HTML5 'required' attribute. However, industrious users
        # or those using older browsers could easily bypass :)
                
        # Snap off the base user parameters
        email = user.get('email', '').strip()
        password = user.get('password', None)
        
        try:
            user = self.db.newuser.new(password=password, email=email)
            user.setsignupinfo(userrec)
            self.db.newuser.put(user)
        except Exception, e:
            self.notify('There was a problem creating your account: %s'%e, error=True)
        else:
            self.simple(content='''Your request for a new account (%s) is being processed. You will be notified via email when it is approved.'''%(user.email))

    @View.add_matcher(r'^/users/queue/$', view='Users', name='queue')    
    def queue(self, action=None, name=None, **kwargs):
        self.template = '/pages/users.queue'    
        self.title = 'Account requests'

        actions = kwargs.pop('actions', {})
        groups = kwargs.pop('groups', {})

        self.ctxt['actions'] = actions

        if self.request_method == 'post':
            # I don't care for this format, but it's a limitation of
            # HTML input elements. 
            reject = set(k for k,v in actions.items() if v=='reject')
            approve = set(k for k,v in actions.items() if v=='approve')
            rejected = []
            approved = []
            if reject:
                rejected = self.db.newuser.reject(reject)
            if approve:
                approved = self.db.newuser.approve(approve)

            # Add the users to all the requested groups.
            # ... do this better in the future.
            for user in approved:
                g = filter(None, groups.get(user.name, []))
                g = emen2.util.listops.check_iterable(g)

                for group in self.db.group.get(g):
                    group.adduser(user.name)
                    self.db.group.put(group)

        queue = self.db.newuser.get(self.db.newuser.filter(None))
        self.ctxt['queue'] = queue

        groupnames = self.db.group.filter(None)
        groupnames -= set(['anon', 'authenticated'])
        self.ctxt['groups'] = self.db.group.get(groupnames)
        self.ctxt['groups_default'] = set(['create'])



__version__ = "$Revision: 1.29 $".split(":")[1][:-1].strip()



# $Id: config.py,v 1.169 2013/05/08 04:51:38 irees Exp $
"""This module manages EMEN2 configurations and options.

Methods:
    get_filename(package, resource)
    defaults()
    get(key, default=None)
    set(key, value)

Classes:
    DBOptions

"""

import os
import sys
import glob
import imp

import jsonrpc.jsonutil

# EMEN2 imports
# NOTHING else should import emen2.db.globalns.
# It is a PRIVATE module!
import emen2.db.globalns

# Note:
#     Be very careful about importing EMEN2 modules here!
#     This module is loaded by many others, it can create circular
#     dependencies very easily!

basestring = (str, unicode)


##### Mako template lookup #####

import mako
import mako.lookup

class AddExtLookup(mako.lookup.TemplateLookup):
    """This is a slightly modified TemplateLookup that
     adds '.mako' extension to all template names.

    Extends TemplateLookup methods:
        get_template        Adds '.mako' to filenames
        render_template        ""

    """
    def get_template(self, uri):
        return super(AddExtLookup, self).get_template('%s.mako'%uri)

    def render_template(self, name, ctxt):
        tmpl = self.get_template(name)
        return tmpl.render(**ctxt)

# Mako Template Loader
# Turn on HTML escaping by default. Use ${variable | n} to disable escaping.

# todo: fix jsonrpc, to escape forward slashes -- until then..
# ["from jsonrpc.jsonutil import encode as jsonencode"]

templates = AddExtLookup(
    input_encoding='utf-8', 
    imports=['from emen2.db.util import jsonencode'],
    default_filters=['h'],
    )



##### Config methods #####

def get_filename(package, resource=None):
    """Get the absolute path to a file inside a given Python package"""
    d = sys.modules[package].__file__
    if resource:
        d = os.path.dirname(d)
        d = os.path.abspath(d)
        return os.path.join(d, resource)
    return d

def get(key, default=None, rv=True):
    """Get a configuration value.

    :param key: Configuration key
    :keyword default: Default value if key is not found
    :return: Configuration value

    """
    result = Config.globalns.watch(key, default=default)
    if rv:
        result = result.get()
    return result

# This will eventually help lock
# the configuration for setting
def set(key, value):
    """Set a configuration value.

    :param key: Configuration key
    :param value: Configuration value

    """
    raise NotImplementedError, "Soon."


##### Extensions #####

def load_exts():
    for ext in Config.globalns.extensions.exts:
        load_ext(ext)

def load_views():
    for ext in Config.globalns.extensions.exts:
        load_view(ext)

def load_jsons(cb=None, *args, **kwargs):
    for ext in Config.globalns.extensions.exts:
        load_json(ext, cb=cb, *args, **kwargs)

def load_ext(ext):
    modulename = 'emen2.exts.%s'%ext
    # print "Loading extension...", modulename
    if modulename in sys.modules:
        # print "%s already loaded"%modulename
        return
    paths = list(Config.globalns.paths.exts)
    module = imp.find_module(ext, paths)
    ret = imp.load_module(ext, *module)
    # Extensions may have an optional "templates" directory,
    # which will be added to the template search path.
    templates.directories.insert(0, os.path.join(module[1], 'templates'))
    return ret

def load_view(ext):
    # Extensions may have an optional "views" module.
    modulename = 'emen2.exts.%s.views'%ext
    # print "Loading views...", modulename
    if modulename in sys.modules:
        # print "%s already loaded"%modulename
        return
    paths = list(Config.globalns.paths.exts)
    module = imp.find_module(ext, paths)
    path = module[1]
    try:
        viewmodule = imp.find_module('views', [path])
    except ImportError, e:
        viewmodule = None
    if viewmodule:
        imp.load_module(modulename, *viewmodule)


def load_json(ext, cb=None, *args, **kwargs):
    path = resolve_ext(ext)
    if not cb:
        return
    for j in sorted(glob.glob(os.path.join(path, 'json', '*.json'))):
        cb(j, *args, **kwargs)

def resolve_ext(ext):
    paths = list(Config.globalns.paths.exts)
    return imp.find_module(ext, paths)[1]



##### Configuration loader #####

class Config(object):
    globalns = emen2.db.globalns.GlobalNamespace()

    def load_file(self, fn):
        '''Load a single configuration file

        :param fn: the filename of the configuration file'''
        self.globalns.from_file(fn)

    def load_data(self, *args, **data):
        '''Load configuration variables into the namespace'''
        if args:
            for dct in args:
                self.load_data(**dct)

        for key, value in data.items():
            # This needs to use setattr() so @properties will work.
            setattr(self.globalns, key, value)

    def require_variable(self, var_name, value=None, err_msg=None):
        '''Assert that a certain variable has been loaded

        :param var_name: the variable to be checked
        :param value: the value it should have, if this is None, the value is ignored
        :param err_msg: the error message to be displayed if the variable is not found
        :raises ValueError: if the variable is not found'''

        #NOTE: if we want None to be a valid config option, this must change
        if value is not None and self.globalns.getattr(var_name, value) != value:
            raise ValueError(err_msg)
        elif self.globalns.getattr(var_name) is None:
            raise ValueError(err_msg)
        else:
            return True



##### Default OptionParser #####
# This has been converted to Twisted usage.Options parser
# to work with "twistd".

from twisted.python import usage

class DBOptions(usage.Options):
    """Base database options."""

    optFlags = [
        ['quiet', None, 'Quiet'],
        ['debug', None, 'Print debug'],
        ['version', None, 'EMEN2 Version'],
        ['create', None, 'Create new database environment'],
        ['nosnapshot', None, 'Disable Berkeley DB Multiversion Concurrency Control (Snapshot)']
    ]

    optParameters = [
        ['home', 'h', None, 'EMEN2 database environment directory'],
        ['ext', 'e', None, 'Add extension; can be comma-separated.'],
        ['loglevel', 'l', None, '']
    ]

    def opt_configfile(self, file_):
        self.setdefault('configfile', []).append(file_)

    opt_c = opt_configfile

    def postProcess(self):
        ## note that for optFlags self[option_name] is 1 if the option is given and 0 otherwise
        ##     this converts those values into the appropriate bools
        # these default to True:
        for option_name in ['create', 'quiet', 'debug', 'version']:
            self[option_name] = bool(self[option_name])

        # these default to False:
        for option_name in ['nosnapshot']:
            self[option_name] = not bool(self[option_name])

    def load_config(self):
        # Do additional processing during configuration loading
        pass


class UsageParser(object):

    def __init__(self, optclass=None, options=None):
        # Use the default DBOptions if none is provided
        if not optclass:
            optclass = DBOptions

        if not options:
            options = optclass()
            options.parseOptions()

        self.options = options
        self.config = Config()
        self.load_config()


    def load_config(self, **kw):
        if self.config.globalns.getattr('CONFIG_LOADED', False):
            return

        # Eventually 'with' will unlock/lock the globalns
        # with globalns:
        self._load_config(**kw)
        self.config.globalns.CONFIG_LOADED = True


    def _load_config(self, **kw):
        # Set EMEN2DBHOME from the options or environment variable.
        h = self.options.get('home', os.getenv("EMEN2DBHOME"))

        # print "EMEN2 config loader: %s"%h
        self.config.load_data(EMEN2DBHOME=h)

        # Load the base configuration.
        self.config.load_file(get_filename('emen2', 'db/config.base.json'))

        # Load other specified config files
        for f in self.options.get('configfile', []):
            self.config.load_file(f)

        # EMEN2DBHOME must have been specified in either -h, $EMEN2DBHOME,
        # or set a configuration file.
        self.config.require_variable(
            'EMEN2DBHOME',
            None,
            err_msg="You must specify an EMEN2 database environment, either using the -h (--home) argument or the environment variable $EMEN2DBHOME")
        if h is None:
            h = self.config.globalns.getattr('EMEN2DBHOME', h)

        # Load any config file in EMEN2DBHOME
        self.config.load_file(os.path.join(h, "config.json"))

        # Set default log levels
        log_level = self.config.globalns.getattr('log_level', 'INFO')
        if self.options['quiet']:
            log_level = 'ERROR'
        elif self.options['debug']:
            log_level = 'DEBUG'
        elif self.options['loglevel']:
            log_level = self.options['loglevel']
        self.config.load_data(log_level=log_level)

        # Make sure paths to log files exist
        if not os.path.exists(self.config.globalns.paths.log):
            os.makedirs(self.config.globalns.paths.log)

        # Extend the python module path
        # if getattr(self.config.globalns.paths, 'PYTHONPATH', []):
        #    pp = self.config.globalns.paths.PYTHONPATH
        #    if not hasattr(pp, '__iter__'):
        #        pp = [pp]
        #    sys.path.extend(pp)

        # EXTPATHS points to directories containing emen2 ext modules.
        # This will be used with imp.find_module(ext, self.config.globalns.paths.exts)
        self.config.globalns.paths.exts.append(get_filename('emen2', 'exts'))
        if os.getenv('EMEN2EXTPATH'):
            for path in filter(None, os.getenv('EMEN2EXTPATH','').split(":")):
                self.config.globalns.paths.exts.append(path)

        self.config.globalns.paths.exts.append(os.path.join(h, 'exts'))


        # Add the extensions, including the 'base' extension
        exts = self.options.get('ext')
        if exts:
            exts = exts.split(',')
            if 'base' not in exts:
                exts.insert(0,'base')
            exts.extend(self.config.globalns.extensions.exts)
            self.config.globalns.extensions.exts = exts

        # Create new database?
        self.config.globalns.params.create = self.options['create']

        # Enable root user?
        # self.config.globalns.ENABLEROOT = self.values.enableroot or False

        # Do anything defined by the usage.Options class
        self.options.load_config()

        # Tell the logger that we're initialized!
        import emen2.db.log
        emen2.db.log.logger.init()


__version__ = "$Revision: 1.169 $".split(":")[1][:-1].strip()



# $Id: record.py,v 1.94 2013/05/08 04:51:38 irees Exp $
"""Record DBOs

Classes:
    Record
    RecordDB

"""

import collections
import copy

# EMEN2 imports
import emen2.db.exceptions
import emen2.db.dataobject
import emen2.db.recorddef
import emen2.util.listops as listops


class Record(emen2.db.dataobject.PermissionsDBObject):
    """Database Record.

    Provides the following additional attributes:
        rectype, history, comments

    This class represents a single database record. In a sense this is an
    instance of a particular RecordDef, however, note that it is not required
    to have a value for every field described in the RecordDef, although this
    will usually be the case. This class is a subclass of PermissionsDBObject
    (and BaseDBObject); see these classes for additinal documentation.

    The RecordDef name is stored in the rectype attribute. Currently, this
    cannot be changed after a Record is created, even by admins. However, this
    functionality may be provided at some point in the future.

    Unlike most other DBOs, Records allow arbitrary attributes
    as long as they are valid EMEN2 Parameters. These are stored in the params
    attribute, which is a dictionary with parameter names as keys.  The params
    attribute is effectively private and not exported. Instead, it is part of
    the mapping interface. Items can be set with __setitem__
    (record['parameter'] = Value) or through an update(). When an item is
    exported (e.g. JSON), the contents of param are in the regular dictionary
    of attributes. Changes to these parameters are always logged in the history
    log, described below, and will always trigger an update to the
    modification time.

    Records contain an integrated log of all changes over the entire history
    of the record. In a sense, as in a physical lab notebook, an original value
    can never be changed, only superceded. This log is stored in the history
    attribute, which is a list containing a tuple entry for every change made,
    in the following format:

    0
        User
    1
        Time of change
    2
        Parameter changed
    3
        Previous parameter value

    The history log is immutable, even to admins, and is updated when the item
    is committed. From a database standpoint, this is rather odd behavior. Such
    tasks would generally be handled with an audit log of some sort. However,
    in this case, as an electronic representation of a Scientific lab notebook,
    it is absolutely necessary that all historical values are permanently
    preserved for any field, and there is no particular reason to store this
    information in a separate file. Generally speaking, such changes should be
    infrequent. When a value is edited, the user interface should generally
    prompt the user for a comment describing the reason for the change. If
    provided, the comment and the history log item will have the same timestamp.

    Users also can store free-form textual comments in the comments attribute,
    either by setting the comments key (item['comments'] = 'Test') or through
    the  addcomment() method. Comments will stored as plain text, and usually
    displayed with Markdown-type formatting applied. Like the history log,
    comments are immutable once set, even to admins. Additionally, parameters
    can be updated inside of a comment using the RecordDef view syntax:
        $$parameter="value"

    Each new comment is added to the comments list as a tuple with the format:

    0
        User
    1
        Time of comment
    2
        Comment text


    The following methods are overridden:

    init
                Init the rectype, comments, and history
    keys
                Add parameter keys

    :attr history: History log
    :attr comments: Comments log
    :attr rectype: Associated RecordDef

    """

    #: Attributes readable by the user
    attr_public = emen2.db.dataobject.PermissionsDBObject.attr_public | set(['comments', 'history', 'rectype'])

    #: Attributes that cannot be edited directly by the user
    attr_protected = emen2.db.dataobject.PermissionsDBObject.attr_protected | set(['comments', 'history'])

    #: Attributes required for validation
    attr_required = set(['rectype'])

    #: The id of the record, for backwards compatibility only
    recid = property(lambda s:s.name)

    def init(self, d):
        # Call PermissionsDBObject init
        super(Record, self).init(d)

        # rectype is required
        # Access to RecordDef is checked during validation
        self.__dict__['rectype'] = None

        # comments, history, and other param values
        self.__dict__['comments'] = []
        self.__dict__['history'] = []
        self.__dict__['params'] = {}
        
        # Records are initialized with these two parameters....
        # ... edit: turning off for now.
        # self.__dict__['params']['date_occurred'] = self.__dict__['creationtime']
        # self.__dict__['params']['performed_by'] = self.__dict__['creator']

    def __repr__(self):
        return "<%s %s, %s at %x>" % (self.__class__.__name__, self.name, self.rectype, id(self))

    ##### Setters #####

    def _set_rectype(self, key, value):
        if not self.isnew():
            self.error("Cannot change rectype")
        self.__dict__['rectype'] = unicode(value)
        return set(['rectype'])
        
    def _set_comments(self, key, value):
        """Bind record['comments'] setter"""
        return self.addcomment(value)

    # in Record, params not in self.attr_public are put in self.params{}.
    def _setoob(self, key, value):
        """Set a parameter value."""
        # Check write permission
        if not self.writable():
            msg = "Insufficient permissions to change param %s"%key
            self.error(msg, e=emen2.db.exceptions.SecurityError)

        # No change
        if self.params.get(key) == value:
            return set()

        self._addhistory(key)
        # Really set the value
        self.params[key] = value
        return set([key])

    ##### Tweaks to mapping methods #####

    def __getitem__(self, key, default=None):
        """Default behavior is similar to .get: return None as default"""
        if key in self.attr_public:
            return getattr(self, key, default)
        else:
            return self.params.get(key, default)

    def keys(self):
        """All retrievable keys for this record"""
        return self.params.keys() + list(self.attr_public)

    def items(self):
        return [(k,self[k]) for k in self.keys()]

    def paramkeys(self):
        return self.params.keys()

    ##### Comments and history #####

    def _addhistory(self, param):
        """Add an entry to the history log."""
        # Changes aren't logged on uncommitted records
        if self.isnew():
            return
        if not param:
            raise Exception, "Unable to add item to history log"
        self.history.append((unicode(self._ctx.username), unicode(self._ctx.utcnow), unicode(param), self.params.get(param)))

    def addcomment(self, value):
        """Add a comment. Any $$param="value" comments will be parsed and set as values.

        :param value: The comment to be added
        """
        if not self.commentable():
            self.error('Insufficient permissions to add comment', e=emen2.db.exceptions.SecurityError)

        if not value:
            return set()

        cp = set()
        if value == None:
            return set()

        # Grumble...
        newcomments = []
        existing = [i[2] for i in self.comments]
        if not hasattr(value, "__iter__"):
            value = [value]
        for c in value:
            if hasattr(c, "__iter__"):
                c = c[-1]
            if c and c not in existing:
                newcomments.append(unicode(c))

        # newcomments2 = []
        # updvalues = {}
        for value in newcomments:
            d = {}
            if not value.startswith("LOG"): # legacy fix..
                d = emen2.db.recorddef.parseparmvalues(value)[1]

            if d.has_key("comments"):
                # Always abort
                self.error("Cannot set comments inside a comment", warning=False)

            # Now update the values of any embedded params
            for i,j in d.items():
                cp |= self.__setitem__(i, j)

            # Store the comment string itself
            self.comments.append((unicode(self._ctx.username), unicode(self._ctx.utcnow), unicode(value)))
            cp.add('comments')

        return cp

    ##### Validation #####

    def validate(self):
        """Validate the record before committing."""
        # Cut out any None's
        # The rest of the parameters are validated 
        # when they are set or updated.
        pitems = self.params.items()
        for k,v in pitems:
            if not v and v != 0 and v != False:
                del self.params[k]

        # Check the rectype and any required parameters
        # (Check the cache for the recorddef)
        cachekey = ('recorddef', self.rectype)
        hit, rd = self._ctx.cache.check(cachekey)

        if not self.rectype:
            self.error('Protocol required')

        if not hit:
            try:
                rd = self._ctx.db.recorddef.get(self.rectype, filt=False)
            except KeyError:
                self.error('No such protocol: %s' % self.rectype)
            self._ctx.cache.store(cachekey, rd)

        # This does rely somewhat on validators returning None if empty..
        for param in rd.paramsR:
            if self.get(param) is None:
                self.error("Required parameter: %s"%(param))

        self.__dict__['rectype'] = unicode(rd.name)


__version__ = "$Revision: 1.94 $".split(":")[1][:-1].strip()



# $Id: paramdef.py,v 1.47 2013/05/01 16:32:12 irees Exp $
"""Parameter DBOs

Classes:
    ParamDef
    ParamDefDB

"""

import functools
import time

# EMEN2 imports
import emen2.util.listops
import emen2.db.dataobject
import emen2.db.magnitude


class ParamDef(emen2.db.dataobject.BaseDBObject):
    """Parameters.
    
    Provides the following attributes:
        desc_short, desc_long, vartype, choices, defaultunits, property, 
        indexed, iter, immutable, property, indexed, controlhint
    
    This class repesents a particular piece of data, either an attribute on a
    DBO, or a parameter on a Record. Please be aware that several
    attributes are effectively immutable, and cannot be easily changed after
    a ParamDef is created because it would change the meaning of existing
    recorded data or index formats. This is a subclass of BaseDBObject; see 
    that class for additional documentation.
    
    Each ParamDef has a vartype (data type), which provides a class for 
    displaying and validating the attribute/parameter value. Vartypes are
    defined as Vartype subclasses; see the vartypes module. Vartypes are usually 
    able to handle loosely formatted data and clean it up (validate), or raise 
    a ValidationError exception if it cannot be done, or if a value is invalid.
    The validation methods are usually called whenever an attribute/parameter
    is set on an item.
    
    The iter attribute is used in conjunction with the vartype to specify
    whether the ParamDef stores a single value, or a list of values. Most
    vartypes allow iter, but some do not. Check the specific Vartype class.
    Vartypes are usually good at validating a single value into an iterable
    value, but not the reverse. The vartype cannot be easily changed after 
    creation; to do so requires running a migration script, which could require
    changing or discarding data, and likely rebuilding indexes.
    
    When ParamDefs are displayed in certain contexts, such as being rendered
    in a view or as a table header, the desc_short and desc_long attributes
    are used to provide short and long descriptions, respectively, of the
    ParamDef and its intended uses and requirements. The desc_short attribute
    is required, desc_long is optional. Because they do not change the actual
    values of recorded data, they can be edited (within reason) after creation.
    
    The choices attribute can be used to provide a default list of choices
    when entering values. These are provided as suggestions for consistency
    among users, and are usually combined with query results to give additional
    common choices. However, for vartype 'choice', this attribute is enforced
    as the only allowed values. Choices can be edited after creation.
    
    Editing widgets are usually determined by the vartype (this choice is
    always up to that particular user interface.) However, some vartypes might
    be edited in several different fashions, and a particular editing control
    might be more effective than the default. In these cases, the controlhint
    attribute can be used to describe a different editing widget. This is
    totally implementation specific; it is only provided here as a string with
    no additional validation. The controlhint can be changed after creation.
    
    If a ParamDef represents a particular measurable physical property (length,
    volume, mass, pH, etc.) the property attribute can be set. Properties are
    defined as subclasses of Property; see the properties module for available
    classes. Properties provide additional methods for understanding units,
    converting between different units or combinations of units (e.g. meters
    to kilometers, degrees Celsius to Fahrenheit, etc.) and listing what units
    are valid for a particular physical property. If a property is set, you can
    usually set a parameter with units, such as:
    
        record['temperature_ambient'] = '72F'

    If the temperature_ambient ParamDef has the 'temperature' property and
    'degC' as the defaultunits, the 72F would be converted to 22.2 degrees C,
    and stored as float(22.2). The defaultunits attribute selects which of the 
    acceptable units defined by the Property will be the default for this
    ParamDef. See the Property class for additioal details. The property and 
    defaultunits cannot be easily changed after creation; if the units change, 
    or the change will require a change in stored Record values or  indexes, 
    it can only be changed by running a migration script.
        
    The indexed attribute will determine if a ParamDef is indexed or not (if
    the DBO type supports indexing.) If it will not be important
    to run frequent queries on the parameter, or if the parameter changes very
    frequently, you might turn off indexing. This cannot be easily changed
    after creation; doing so requires rebuilding the index.
    
    Finally, the immutable attribute is used to lock an parameter value. If 
    set, no ParamDef of this type can be edited. This can be turned on/off after
    creation, within reason.
        
    The following methods are overridden:
        init        Initialize ParamDef attributes
        validate    Check the vartype and property are valid

    :attr desc_short: Short description of ParamDef
    :attr desc_long: Long description of ParamDef and indended uses
    :attr vartype: Data type
    :attr choices: A list of suggested (or restricted) choices
    :attr defaultunits: Default units for physical property
    :attr property: Physical property
    :attr indexed: Values are indexed
    :attr iter: Values can be iterable
    :attr immutable: ParamDef is locked
    :attr controlhint: Hint for user-interface widget

    """
    attr_public = emen2.db.dataobject.BaseDBObject.attr_public | set(['immutable', 
        'iter', 'desc_long', 'desc_short', 'choices', 'vartype',
        'defaultunits', 'property', 'indexed', 'controlhint'])
    attr_required = set(['vartype'])


    def init(self, d):

        # Variable data type. List of valid types in the module global 'vartypes'
        self.__dict__['vartype'] = None

        # This is a very short description for use in forms
        self.__dict__['desc_short'] = None

        # A complete description of the meaning of this variable
        self.__dict__['desc_long'] = None

        # Physical property represented by this field, List in 'properties'
        self.__dict__['property'] = None

        # Default units (optional)
        self.__dict__['defaultunits'] = None

        # choices for choice and string vartypes, a tuple
        self.__dict__['choices'] = []

        # Immutable
        self.__dict__['immutable'] = False
        
        # Iterable. This can be False, True (list), list, set, dict.
        self.__dict__['iter'] = False

        # turn indexing on/off, if vartype allows for it
        self.__dict__['indexed'] = True
        
        # Widget hint
        self.__dict__['controlhint'] = None

    #################################
    # Setters
    #################################

    # ParamDef does so much validation for other items, 
    # so everything is checked....
    # Several values can only be changed by administrators.

    def _set_choices(self, key, value):
        value = emen2.util.listops.check_iterable(value)
        value = filter(None, [unicode(i) for i in value]) or None
        return self._set(key, value, self.isowner())

    def _set_desc_short(self, key, value):
        return self._set(key, unicode(value or self.name), self.isowner())

    def _set_desc_long(self, key, value):
        return self._set(key, unicode(value or ''), self.isowner())

    # Only admin can change defaultunits/immutable/indexed/vartype.
    # This should still generate lots of warnings.
    def _set_immutable(self, key, value):
        return self._set(key, bool(value), self._ctx.checkadmin())

    def _set_iter(self, key, value):
        return self._set(key, bool(value), self._ctx.checkadmin())

    def _set_indexed(self, key, value):
        return self._set(key, bool(value), self._ctx.checkadmin())

    def _set_controlhint(self, key, value):
        if value != None:
            value = unicode(value)
        value = value or None
        return self._set(key, value)
                
    # These can't be changed, it would disrupt the meaning of existing Records.
    def _set_vartype(self, key, value):
        if not self.isnew():
            self.error("Cannot change vartype from %s to %s."%(self.vartype, value))

        value = unicode(value or '') or None

        if value not in emen2.db.vartypes.Vartype.registered:
            self.error("Invalid vartype: %s"%value)

        return self._set(key, value)

    def _set_property(self, key, value):
        if not self.isnew():
            self.error("Cannot change property from %s to %s."%(self.property, value))

        value = unicode(value or '')
        if value in ['None', None, '']:
            value = None

        # Allow for unsetting
        if value != None and value not in emen2.db.properties.Property.registered:
            self.error("Invalid property: %s"%value)

        return self._set('property', value)

    def _set_defaultunits(self, key, value):
        if not self.isnew():
            self.error("Cannot change defaultunits from %s to %s."%(self.defaultunits, value))

        value = unicode(value or '') or None
        value = emen2.db.properties.equivs.get(value, value)
        return self._set('defaultunits', value)

    def validate(self):
        if not self.vartype:
            self.error("Vartype required")
        if self.vartype not in emen2.db.vartypes.Vartype.registered:
            self.error("Vartype %r is not a valid vartype" % self.vartype)
    #     try:
    #         prop = emen2.db.properties.Property.get_property(self.property)
    #     except KeyError:
    #         self.error("Cannot set defaultunits without a property!")
    #    m = emen2.db.magnitude.mg(0, value)
    #    # self.error("Invalid units: %s"%value)
    #     if value not in prop.units:
    #         self.error("Invalid defaultunits %s for property %s. 
    #             Allowed: %s"%(value, self.property, ", ".join(prop.units)))

__version__ = "$Revision: 1.47 $".split(":")[1][:-1].strip()


import twisted.python.log
import twisted.application

import emen2.db.log
import emen2.db.config
import emen2.web.server

Options = emen2.web.server.WebServerOptions


def logger():
    emen2.db.log.logger.start()
    return emen2.db.log.logger.emit

def makeService(options):
    # Load the configuration
    import emen2.db.config
    emen2.db.config.UsageParser(options=options)

    # Start the service
    s = twisted.application.service.MultiService()
    server = emen2.web.server.EMEN2WebServer(options)
    server.start(service=s)    
    return s
    
def run_twistd():
    # This is a (somewhat awful hack) based on twistd.

    # Twistd plugins have great difficulty being packaged
    # by distutils, pip, etc., and there is no
    # current workaround. 
    
    # This function basically recreates twistd,
    # and inserts the EMEN2 Service as the only plugin.
    # This gets around having to have twisted/plugins/emen2_plugin.py
    # in a directory without an __init__.py

    # Run:
    # python -m emen2.tap [twistd options] emen2 [emen2 options]

    from twisted.application.service import ServiceMaker
    from twisted.application import app
    from twisted.python.runtime import platformType
    from twisted import plugin

    #### From twisted/plugins/emen2_plugin.py #####
    EMEN2Server = ServiceMaker(
        "EMEN2",
        "emen2.tap",
        ("EMEN2 server"),
        "emen2")

    #### Copied from twisted/scripts/twistd.py #####
    if platformType == "win32":
        from twisted.scripts._twistw import ServerOptions, \
            WindowsApplicationRunner as _SomeApplicationRunner
    else:
        from twisted.scripts._twistd_unix import ServerOptions, \
            UnixApplicationRunner as _SomeApplicationRunner

    #Create an altered ServerOptions that only runs EMEN2.
    class TestOptions(ServerOptions):
        def _getPlugins(self, interface, package=None):
            f = plugin.getPlugins(interface, package=None)
            return list(f) + [EMEN2Server]

    def runApp(config):
        _SomeApplicationRunner(config).run()
    
    app.run(runApp, TestOptions)
    
    

if __name__ == "__main__":
    run_twistd()
    



# $Id: resource.py,v 1.49 2013/01/04 20:52:21 irees Exp $

import re
import time
import cgi
import traceback
import collections
import functools
import Queue
import itertools

import twisted.internet
import twisted.web.static
import twisted.web.resource
import twisted.python.failure

import jsonrpc.jsonutil
import jsonrpc.server

import mako.exceptions

# emen2 imports
import emen2.db.config
import emen2.db.log
import emen2.db.exceptions
import emen2.db.handlers
import emen2.util.listops
import emen2.web.events
import emen2.web.routing
import emen2.web.responsecodes
import emen2.web.server

##### Base EMEN2 Resource #####

# Special thanks to this Dave Peticolas's blog post for
# helping sort out my deferred cancel/callback/errback situation:
# http://krondo.com/?p=2601

class RoutedResource(object):
    
    ##### Routing registration methods #####
    routing = emen2.web.routing
    isLeaf = True

    def render(self, request):
        return "No content."

    @classmethod
    def add_matcher(cls, *matchers, **kwargs):
        '''Decorator used to add a matcher to an already existing class

        Named groups in matcher get passed as keyword arguments
        Other groups in matcher get passed as positional arguments
        Nothing else gets passed

        write=True will hint to the database that writes will occur.
        This may do things like disable snapshot transactions.
        '''
        if not matchers:
            raise ValueError, 'A view must have at least one matcher'

        # Default name (this is usually the method name)
        def check_name(name):
            return 'main' if name.lower() == 'init' else name

        # Inner decorator method
        def inner(func):
            name = kwargs.pop('name', check_name(func.__name__))
            view = kwargs.pop('view', None)
            write = kwargs.pop('write', False)
            matcherinfo = getattr(func, 'matcherinfo', [])
            for count, m in enumerate(matchers):
                if count>0:
                    name='%s/%s'%(name, count)
                matcherinfo.append((m, name, view, write))

            # save all matchers to the function
            func.matcherinfo = matcherinfo
            return func

        return inner

    @classmethod
    def register(self, cls):
        '''Register a View and connect it to a URL.
        - Multiple regular expressions can be registered per method
        - This also registers urls defined by the add_matcher decorator. In this case, the sub view name will default to the method name.
        - These can be reversed with self.ctxt.reverse('ClassName/alt1', param1='asd') and such
        '''
        # Register mtchers produced by the add_matcher decorator
        for func in cls.__dict__.values():
            if not callable(func): continue

            for matcher in getattr(func, 'matcherinfo', []):
                matcher, name, view, write = matcher
                view = view or cls.__name__
                name = '%s/%s'%(view, name)
                # These will be set as attributes on the Route instance
                with emen2.web.routing._Router().route(name=name, matcher=matcher, cls=cls, method=func, write=write) as url:
                    pass

        return cls

    slots = collections.defaultdict(list)
    @classmethod
    def provides(cls, slot):
        '''Decorate a method to indicate that the method provides a certain functionality'''
        def _inner(view):
            cls.slots[slot].append(functools.partial(cls, init=view))
            return view
        return _inner

    @classmethod
    def require(cls, slot):
        '''Use to get a view with a desired functionality'''
        if slot in cls.slots:
            return cls.slots[slot][-1]
        else: raise ValueError, "No such slot"


class FixedArgsResource(object):
    """Better handling of request args than Twisted's Resource."""
    
    def render(self, request):
        return "No content."
    
    ##### Process request arguments #####

    def parse_args(self, request):
        # Massage the post and querystring arguments
        # ... PUT requests, and twistd's handling of POST is broken
        args = self._parse_content(request)

        # Twisted provides all args as lists.
        # If we only got one value, pop it
        for k, v in args.items():
            if len(v) == 1:
                v = v[0]
            args[k] = v

        # Unicode.. grumble.. Web forms will submit UTF-8 encoded data; decode that
        args = self._parse_coerce_unicode(args)

        # HTTP arguments with '.' will be turned into dicts, e.g. 'child.key' -> child['key']
        args = self._parse_args_dict(args)

        # Redirect...?
        self._redirect = args.pop('_redirect', None)

        return args

    def _parse_content(self, request):
        # Fixes file uploads.
        files = []
        args = request.args

        if request.method == "PUT":
            # The param?..
            f = emen2.db.handlers.BinaryHandler.get_handler(
                filename=request.getHeader('x-file-name'),
                param=request.getHeader('x-file-param'),
                fileobj=request.content
                )
            files.append(f)

        elif request.method == "POST":
            # Fix Twisted's broken handling of multipart/form-data file uploads
            headers = request.getAllHeaders()
            img = cgi.FieldStorage(
                fp = request.content,
                headers = headers,
                keep_blank_values = True,
                environ = {
                    'REQUEST_METHOD':'POST',
                    'CONTENT_TYPE': headers.get('content-type'),
                }
            )

            # Rebuild the request args.
            # args = {}
            for param in img:
                # img.getlist() only returns the values.
                newvalues = []
                values = img[param]
                # grumble.. hasattr(__iter__) doesn't work.
                if not isinstance(values, list):
                    values = [values]
                for value in values:
                    if getattr(value, 'filename', None):
                        f = emen2.db.handlers.BinaryHandler.get_handler(
                            param=param,
                            filename=value.filename,
                            fileobj=value.file or value.value
                            )
                        files.append(f)
                    else:
                        newvalues.append(value.value)
                args[param] = newvalues

        # Fix Unicode filename... Arrrgghghh..
        for f in files:
            f.filename = f.filename.decode('utf-8')

        # Make available to Views...
        self.request_files = files
        return args

    def _parse_args_dict(self, args):
        # Break keys with '.' into child dictionaries, recursively.
        # {'root': 1, 'child.key2': 2, 'child.key1.subkey1': 3, 'child.key1.subkey2':4}
        # ->
        # {'root': 1, 'child': {'key2': 2, key1: {'subkey1': 3, 'subkey2': 4}}}
        test = {}
        newargs = {}
        # Sort by key length so child dictionaries are created in order
        for k,v in sorted(args.items(), key=lambda x:len(x[0])):
            if '.' in k:
                cur = newargs
                s = k.split('.')
                for path in s[:-1]:
                    if not cur.get(path):
                        # Create child dict
                        cur[path] = {}
                    # Step down one level
                    cur = cur[path]
                    parent = path

                # Set the value for the leaf
                cur[s[-1]] = v
            elif ':' in k:
                i, _, j = k.partition(':')
                if not test.get(i):
                    test[i] = {}
                test[i][j] = v

            else:
                newargs[k] = v

        # Transform :-keyed items back into dict
        for k,v in test.items():
            v2 = zip(v.get('keys', []), v.get('values', []))
            newargs[k] = dict(v2)

        return newargs

    def _parse_coerce_unicode(self, args, keyname=''):
        # This is terribly hacky, to deal with various Unicode issues
        # To disable this for a class, override and return {}
        if isinstance(args, unicode):
            return args

        elif keyname == 'filedata':
            # Ignore
            return args

        elif hasattr(args, 'items'):
            newargs = {}
            for k, v in args.items():
                newargs[k] = self._parse_coerce_unicode(v, keyname=k)
            return newargs

        elif hasattr(args, '__iter__'):
            return [unicode(i, 'utf-8') for i in args]

        return unicode(args, 'utf-8')



class EMEN2Resource(RoutedResource, FixedArgsResource):
    """Base resource for EMEN2. 
    
    Handles proper parsing of HTTP request arguments
    into a format that plays well with EMEN2, and contains all the common features
    of setting up the deferreds, threadpool, handling errors, etc.
    """

    def __init__(self, db=None):
        # Database connection, if given.
        self.db = db
        
        # Response headers
        self.headers = {}

        # HTTP Method
        self.request_method = None # request_method

        # Request headers
        self.request_headers = {}

        # Request host
        self.request_host = None
        
        # Request location
        self.request_location = ''
        
        # Redirect...
        self._redirect = None
        
        # Any uploaded files
        self.request_files = []

        # HTTP ETags (cache control)
        self.etag = None

    def __unicode__(self):
        '''Render the resource into a string that can be sent to the client'''
        return unicode(self.get_data())

    def __str__(self):
        '''Render the resource, encoded as UTF-8'''
        return self.get_data().encode('utf-8', 'replace')


    ##### Headers #####

    def set_header(self, name, value):
        '''Set a single header'''
        name = self._normalize_header_name(name)
        self.headers[name] = value
    
    def _normalize_header_name(self, name):
        return '-'.join(x.capitalize() for x in name.split('-'))

    def _set_ctxid(self, request, ctxid):
        request.addCookie("ctxid", ctxid, path='/')

    ##### Resource interface #####

    def get_json(self):
        return None

    def get_data(self):
        return ""
        
    def render(self, request, method=None):
        # The default is to run a supplied method
        # wrapped in a DB transaction from the DBPool.
        method = method or (lambda x:None)

        # Update request details
        self.request_method = request.method.lower()
        self.request_headers = request.getAllHeaders()
        self.request_location = request.path
        self.request_host = request.getClientIP()
        
        # Parse and filter the request arguments
        args = self.parse_args(request)

        # Authentication token -- if supplied as an argument, send it back as a cookie.
        ctxid = args.pop('ctxid', None)
        if ctxid:
            self._set_ctxid(request, str(ctxid))
        ctxid = ctxid or request.getCookie("ctxid")

        # _render will setup the deferred response
        return self._render(request, method, ctxid=ctxid, host=self.request_host, args=args)

    def _render(self, request, method, ctxid=None, host=None, args=None):
        # Setup deferred rendering using a DB transaction.
        t = time.time()

        # Use the EMEN2 DB thread pool.
        deferred = emen2.web.server.pool.rundb(
            self._render_db,
            method,
            ctxid = ctxid,
            host = host,
            args = args)

        # Callbacks
        deferred.addCallback(self.render_cb, request, t=t)
        deferred.addErrback(self.render_eb, request, t=t)
        request.notifyFinish().addErrback(self._request_broken, request, deferred)
        return twisted.web.static.server.NOT_DONE_YET

    def _render_db(self, method, db=None, ctxid=None, host=None, args=None):
        # Render method
        self.db = db

        # Result
        result = None
        
        # Hack to log username and ctxid
        self._log_username = None
        self._log_ctxid = ctxid

        # The DBProxy context manager will open a transaction, and abort
        # on an uncaught exception.
        write = getattr(method, "write", False)
        with self.db._newtxn(write=write):
            # Bind the ctxid/host to the DBProxy
            self.db._setContext(ctxid, host)
            self._log_username = self.db._ctx.username
            
            # Any View init method is run inside the transaction
            self.init()
            result = method(self, **args)
 
             # If the method returns a value, it uses that as the value
            # otherwise, calls str() on the View.
            if result is None:
                result = str(self)

        self.db = None

        return result


    ##### Callbacks #####

    def render_cb(self, result, request, t=0, **_):
        # Render callback -- setup basic headers

        # This is a hack to log the ctxid and username
        request._log_username = self._log_username
        request._log_ctxid = self._log_ctxid
        
        # Filter the headers.
        headers = {}
        headers.update(self.headers)
        headers = dict( (k,v) for k,v in headers.iteritems() if v != None )

        # Redirect if necessary
        if self._redirect:
            headers['Location'] = self._redirect
        if headers.get('Location'):
            request.setResponseCode(303)

        # If X-Ctxid (auth token) was supplied as a header,
        # set the client's cookie. This is is used by login, logout
        # and opening a web browser from desktop clients with ?ctxid as
        # a querystring argument.
        setctxid = headers.pop('X-Ctxid', None)
        if setctxid != None:
            self._set_ctxid(request, setctxid)

        # Set the remaining headers
        [request.setHeader(key, str(headers[key])) for key in headers]

        # Send result to client
        self.render_result(result, request)

    def render_result(self, result, request):
        """Write the result to the client and close the request."""
        # if result is not None:
        length = len(result)
        request.setHeader("Content-Length", length)
        request.write(result)
        
        # Close the request and write to log
        request.finish()

    def render_eb(self, failure, request, t=0, **_):
        # This method accepts either a regular Exception or Twisted Failure
        print failure
        e, data = '', ''
        headers = {}

        # Raise the exception
        try:
            if isinstance(failure, twisted.python.failure.Failure):
                failure.raiseException()
            else:
                raise failure

        # Closed connection error. Nothing to write, and no connection to close.
        except (twisted.internet.defer.CancelledError), e:
            return

        # Expired or invalid session. Remove ctxid and redirect to root.
        except emen2.db.exceptions.SessionError, e:
            data = self.render_error_security(request.uri, e)
            request.addCookie('ctxid', '', path='/')
            emen2.db.log.security(e)

        # Authentication exceptions
        except emen2.db.exceptions.SecurityError, e:
            data = self.render_error_security(request.uri, e)
            emen2.db.log.security(e)

        # HTTP errors
        except emen2.web.responsecodes.HTTPResponseCode, e:
            data = self.render_error_response(request.uri, e)
            emen2.db.log.error(e)

        # General error
        except BaseException, e:
            data = self.render_error(request.uri, e)
            emen2.db.log.error(e)

        # Write the response
        headers.update(getattr(e, 'headers', {}))
        request.setResponseCode(getattr(e, 'code', 500))
        [request.setHeader(k, v) for k,v in headers.items()]
        request.write(data)

        request.finish()


    ##### Error handlers #####

    def render_error(self, location, e):
        # return unicode(emen2.web.routing.execute('Error/main', db=None, error=e, location=location)).encode('utf-8')
        return mako.exceptions.html_error_template().render()

    def render_error_security(self, location, e):
        return unicode(emen2.web.routing.execute('Error/auth', db=None, error=e, location=location)).encode('utf-8')
        # return mako.exceptions.html_error_template().render()

    def render_error_response(self, location, e):
        return unicode(emen2.web.routing.execute('Error/resp', db=None, error=e, location=location)).encode('utf-8')
        # return mako.exceptions.html_error_template().render()

    def _request_broken(self, failure, request, deferred):
        # Cancel the deferred.
        # The errback will be called, but not the callback.
        deferred.cancel()

    def _request_canceller(self, deferred, *args, **kwargs):
        # Do nothing -- my deferreds don't support cancellation at the moment
        pass
        # Create a failure to pass to the errback
        # failure = twisted.python.failure.Failure(exc_value=Exception("Cancelled request"))
        # deferred.errback(failure)








##### XML-RPC and JSON-RPC Resources #####

class XMLRPCResource(object):
    pass


class JSONRPCServerEvents(jsonrpc.server.ServerEvents):
    q = Queue.Queue()

    def processcontent(self, content, request):
        # Get the host from request
        # Get the ctxid from the post or querystring arguments
        ctxid, host = request.getCookie('ctxid'), request.getClientIP()
        ctxid = content.get('ctxid', ctxid)
        if 'params' in content:
            params = content['params']
            if hasattr(params, 'get'):
                ctxid = params.pop('ctxid', ctxid)
        self.ctxid = ctxid
        self.host = request.getClientIP()
        return content

    def callmethod(self, request, rpcrequest, db=None, ctxid=None, **kw):
        # Lookup the method and call
        if not db:
            raise Exception, "No DBProxy"

        # Hack to log username and ctxid
        request._log_username = None
        request._log_ctxid = self.ctxid

        methodresult = None
        if rpcrequest.method.startswith('_'):
            raise emen2.web.responsecodes.ForbiddenError, 'Method not accessible'

        elif rpcrequest.method in db._publicmethods:
            # Start the DB with a write transaction
            # db._starttxn(write=db._checkwrite(rpcrequest.method))
            with db:
                db._setContext(self.ctxid, self.host)
                request._log_username = db._ctx.username
                
                _method = rpcrequest.method.rpartition('.')[2]
                if _method == 'login':
                    rpcrequest.kwargs['host'] = self.host

                methodresult = db._callmethod(rpcrequest.method, rpcrequest.args, rpcrequest.kwargs)
                
                if _method in set(['login', 'logout']):
                    request.addCookie('ctxid', methodresult or '')

        else:
            if rpcrequest.method == 'queue_put':
                methodresult = self.q.put((rpcrequest.args, rpcrequest.kwargs))
            elif rpcrequest.method == 'queue_get':
                methodresult = self.q.get()
            else:
                # Call an event, these can be added and removed through the event registry in emen2.web.events
                #   Useful for setting up views to deliver results/notifications via JSON-RPC
                #     This should work for COMET-style long-polling, but that results in server hangs on exit :)
                e = emen2.web.events.EventRegistry().event('pub.%s'%rpcrequest.method)
                methodresult = e(self.ctxid, request.getClientIP(), db=db, *rpcrequest.args, **rpcrequest.kwargs)
        return methodresult

    def defer(self, method, *a, **kw):
        deferred = emen2.web.server.pool.rundb(method, *a, **kw)
        return deferred

    def log(self, response, txrequest, error=False):
        if error:
            traceback.print_exc()
        else:
            pass





__version__ = "$Revision: 1.49 $".split(":")[1][:-1].strip()



# $Id: group.py,v 1.34 2013/05/01 16:32:12 irees Exp $
"""Group: Represents a group of users, each with certain permissions."""

# EMEN2 imports
import emen2.db.dataobject
import emen2.db.exceptions


class Group(emen2.db.dataobject.PermissionsDBObject):
    """Groups of users.

    Provides the following attributes:
        disabled, displayname, privacy

    Groups are used in conjunction with permissions access control lists to
    share a set of permissions between many records. Some groups are also used
    to provide role-based access, such as administrative rights, the right to
    read published (public) records, the right to create records, etc. Groups
    are only slightly modified from the parent PermissionsDBOBject class; you
    will find additional documentation there.

    When referenced in a Record's groups attribute, the Group's permissions
    will be overlaid on top of the Record permissions. As such, if a user has
    comment level permissions in a group, that user will have comment level
    permissions in any Record that lists that group. All four permissions levels
    will be checked in this way.

    Like a Record, you must have administrative rights in a group to edit that
    groups permissions list.

    The displayname attribute serves a similar purpose to the User displayname
    attribute; it provides a human-formatted description of the group (in some
    cases, groups may have random, arbitrary, or cryptic names.)

    The disabled attribute will disable the Group: it will not be active, and
    Records will not inherit its permissions.

    The following methods are overridden:

        init            Init disabled, displayname, and privacy
        readable        Some special groups are readable by all

    :attr displayname: Human readable display name
    :attr disabled: Group is disabled

    """
    attr_public = emen2.db.dataobject.PermissionsDBObject.attr_public | set(['privacy', 'disabled', 'displayname'])

    def init(self, d):
        super(Group, self).init(d)
        self.__dict__['disabled'] = False
        self.__dict__['displayname'] = None
        self.__dict__['privacy'] = True

    # Groups are readable by anyone.
    def readable(self):
        return True

    # Setters
    def _set_privacy(self, key, value):
        value = int(value)
        if value not in [0,1,2]:
            self.error("Group privacy setting may be 0, 1, or 2.")
        return self._set(key, value, self.isowner())

    def _set_disabled(self, key, value):
        return self._set(key, bool(value), self.isowner())

    def _set_displayname(self, key, value):
        return self._set(key, str(value or self.name), self.isowner())
        

__version__ = "$Revision: 1.34 $".split(":")[1][:-1].strip()


# $Id: download.py,v 1.34 2013/05/14 09:19:12 irees Exp $
import time
import re
import os
import time
import tarfile
import StringIO
import cStringIO

import twisted.web.static

import jsonrpc.jsonutil

# emen2 imports
import emen2.web.resource
from emen2.web.view import View
import emen2.web.responsecodes
import emen2.db.exceptions
import emen2.db.handlers

def renamefile(filename, count=0):
    if not count:
        return filename
    parts = filename.split(".")
    fn = parts[:-1] + ["%s"%count] + parts[-1:]
    return ".".join(fn)

@View.register
class Download(View):

    contentTypes = twisted.web.static.loadMimeTypes()

    contentEncodings = {
            ".gz" : "gzip",
            ".bz2": "bzip2"
            }

    defaultType = 'application/octet-stream'

    @View.add_matcher('^/download/$', name='multi')
    @View.add_matcher('^/download/(?P<bids>[^/]*)/(?P<filename>[^/]*)/$')
    @View.add_matcher('^/binary/(?P<bids>[^/]*/)/$', name='binary')
    def main(self, bids, filename=None, size=None, format=None, q=None, rename=None, tar=None):
        if not hasattr(bids, '__iter__'):
            bids = [bids]

        # Query for BDOs
        if q:
            bdos = self.db.binary.get(q=q)
        else:
            bdos = self.db.binary.get(bids)

        # Found what we needed; close the transaction
        return bdos

    def render_result(self, bdos, request, t=0, **_):
        # Process the returned BDOs into files to send
        size = request.args.get('size')
        format = request.args.get('format', 'jpg')
        rename = request.args.get('rename', False)
        tar = request.args.get('tar', None)
        files = {}
        cache = True
        
        for bdo in bdos:
            name = bdo.get('name', 'none')
            record = bdo.get('record', 'none')
            filename = bdo.get("filename")
            filepath = bdo.get("filepath")
            previewpath = emen2.db.binary.Binary.parse(bdo.get('name')).get('previewpath')
            
            if size:
                # Thumbnail requested
                thumbpath = '%s.%s.%s'%(previewpath, size, format)
                # print "Thumbnail: Checking for...", thumbpath
                if os.access(thumbpath, os.F_OK):
                    # Return the thumbnail
                    files[thumbpath] = '%s.%s.%s'%(filename, size, format)
                else:
                    # Build the thumbnail; return a spinner image
                    cache = False
                    status = emen2.db.handlers.thumbnail_from_binary(bdo, wait=False)
                    files[emen2.db.config.get_filename('emen2', 'web/static/images/handler.%s.png'%status)] = 'handler.%s.png'%status

            elif os.access(filepath, os.F_OK):
                # Found the file
                if rename == 'name':
                    filename = "%s.%s"%(bdo.get('name', 'none').replace('bdo:', ''), filename)
                elif rename == 'record':
                    filename = "%s.%s"%(bdo.get('record', 'none'), filename)
                files[filepath] = filename
            
            else:
                # This will trigger render_eb if the file is not found
                raise IOError, "Could not access file"

        # Check for files that have the same name...
        seen = []
        for k,v in files.items():
            if v in seen:
                files[k] = renamefile(v, count=seen.count(v)+1)
            seen.append(v)
            
        if tar or len(files) > 1:
            return self._transfer_tar(files, request)
        
        return self._transfer_single(files, request, cache=cache)


    ##### Process the result #####

    def _transfer_single(self, files, request, cache=True):
        # Download a single file
        filepath, filename = files.items()[0]
        mimetype, encoding = twisted.web.static.getTypeAndEncoding(filename, self.contentTypes, self.contentEncodings, self.defaultType)

        # If we're saving a gzip, we'll let the browser expand it by setting encoding:gzip.
        if filename[-3:] == ".gz":
            filename = filename[:-3]
            encoding = "gzip"

        fsize = os.stat(filepath).st_size
        f = open(filepath)

        if request.postpath[-1] == "save":
            request.setHeader('Content-Disposition', 'attachment; filename=%s'%filename.encode('utf-8'))

        request.setHeader('Content-Length', str(fsize))
        request.setHeader('Content-Type', mimetype)
        request.setHeader('Content-Encoding', encoding)
        if cache:
            request.setHeader('Cache-Control', 'max-age=86400')

        a = twisted.web.static.NoRangeStaticProducer(request, f)
        a.start()


    def _transfer_tar(self, files, request, cache=False):
        # Download multiple files using TarPipe
        t = emen2.db.database.utcnow()[:10]
        request.setHeader('Content-Disposition', 'attachment; filename=archive-%s.tar'%t)
        request.setHeader('Content-Type', 'application/x-tar')
        request.setHeader('Content-Encoding', 'application/octet-stream')
        a = twisted.web.static.NoRangeStaticProducer(request, TarPipe(files))
        a.start()
        

class TarPipe(object):
    def __init__(self, files={}):
        self.files = files
        self.buffer = cStringIO.StringIO()
        self.tarfile = tarfile.open(mode='w', fileobj=self.buffer)

    def close(self):
        pass

    def _addnextfile(self):
        if not self.files:
            # print "...Closing tarfile"
            self.tarfile.close()
            return
            
        key = self.files.keys()[0]
        filename = self.files.pop(key)

        self.buffer.seek(0)
        self.buffer.truncate(0)
        
        # print "Adding %s: %s... %s files left"%(key, filename, len(self.files))
        self.tarfile.add(key, arcname=filename)
        self.buffer.seek(0)

    def read(self, size=256**2):
        data = self.buffer.read(size)
        if not data:
            self._addnextfile()
            data = self.buffer.read(size)
        return data




__version__ = "$Revision: 1.34 $".split(":")[1][:-1].strip()



# $Id: globalns.py,v 1.100 2012/10/18 09:41:40 irees Exp $
"""Do not use any of these classes or functions directly.
Use the get() and set() functions in emen2.db.config instead.
"""

from __future__ import with_statement

import itertools
import collections

def dict_merge(dct1, dct2):
    commonkeys = set(dct1) & set(dct2)
    newkeys = set(dct2) - commonkeys

    for key in commonkeys:
        olditem = dct1[key]
        newitem = dct2[key]
        if hasattr(newitem, 'items') and hasattr(olditem, 'items'):
            dict_merge(olditem, newitem)
        #elif hasattr(olditem, 'update') and hasattr(newitem, '__iter__'):
        #    olditem.update(newitem)
        #elif hasattr(olditem, 'extend') and hasattr(newitem, '__iter__'):
        #    olditem.extend(newitem)
        else:
            dct1[key] = dct2[key]
    dct1.update( (k,dct2[k]) for k in newkeys )
    return dct1



class _Null:
    pass



class Hier(collections.MutableMapping):
    ####################
    # Class Attributes #
    ####################

    hier = {}

    @classmethod
    def from_dict(cls, dct):
        #cls.init()
        dict_merge(cls.hier, dct)
        return cls()

    #######################
    # Instance Attributes #
    #######################

    def to_dict(self):
        return self._values

    def __init__(self, name=''):
        self._values = self.hier
        self._create = False

        self._name = name

        for segment in self._name.split('.'):
            item = self._values.get(segment)
            if hasattr(item, 'items'):
                self._values = item
            elif name != '':
                raise ValueError('no such item: %r' % self._name)

    def __repr__(self):
        return '%s().from_dict(%r)' % (self.__class__.__name__, self._values)

    def __getattribute__(self, name):
        result = _Null
        try:
            result = object.__getattribute__(self, name)
        except AttributeError:
            if name.startswith('_'): raise
            else:
                result = self.getattr(name, default=_Null, prefix=self._name)
                if isinstance(result, self.__class__):
                    result._create = self._create

        if result is _Null:
            if self._create:
                self.setattr(name, {})
            else:
                raise
        return result

    def __setattr__(self, name, value):
        res = getattr(self.__class__, name, None)
        if name.startswith('_') or hasattr(res, '__set__'):
            object.__setattr__(self, name, value)
        else:
            self.setattr(name, value)

    def getattr(self, name, default=None, prefix=''):
        result = self._values.get(name, _Null)
        if result is _Null and default is not _Null:
            result = default
        elif hasattr(result, 'items'):
            if prefix == '': result = self.__class__(name)
            else: result = self.__class__('%s.%s' % (prefix, name))
        return result

    def __getitem__(self, name):
        result = self.getattr(name, default=_Null)
        if result is _Null:
            raise KeyError('Item %r not found' % name)
        return result

    def setattr(self, name, value, options=None):
        name = name.split('.')

        values = self._values
        for seg in name[:-1]:
            values = values.setdefault(seg, {})
        values[name[-1]] = value

    def __setitem__(self, name, value):
        self._values[name] = value

    def __delitem__(self, name):
        del self._values[name]

    def __iter__(self): return iter(self._values)
    def __len__(self): return len(self._values)



##########################################################
# Direct access is deprecated. Use emen2.db.config: get and set

import re
import collections
import threading
import os
import os.path
import UserDict
import time
import os.path

try: import yaml
except ImportError:
    try: import syck as yaml
    except ImportError:
        yaml = False

try: import json
except ImportError:
    try: import simplejson as json
    except ImportError:
        json = False

from emen2.db import debug
import jsonrpc.jsonutil

inst = lambda x:x()


class GlobalNamespace(Hier):

    def claim(self, name, default=None, validator=None):
        return Claim(self, name, default, validator)

    def watch(self, name, default=None):
        return Watch(self, name, default)

    _events = collections.defaultdict(list)

    @inst
    class paths(object):
        '''A class which stores path configuration variables and adjusts relative pasts to point to the root specified in the root attribute'''
        attrs = []
        __root = ['']

        @property
        def root(self):
            return self.__root[0]

        @root.setter
        def root(self, v):
            self.__root.insert(0, v)

        def get_roots(self):
            return self.__root[:]

        def update(self, path):
            self.root = path

        def __delattr__(self, name):
            try:
                self.attrs.remove(name)
            except ValueError:
                pass
            object.__delattr__(self, name)

        def __setattr__(self, name, value):
            self.attrs.append(name)
            return object.__setattr__(self, name, value)

        def __getattribute__(self, name):
            value = object.__getattribute__(self, name)
            if name.startswith('_') or name in set(['attrs', 'root']):
                return value
            else:
                if isinstance(value, (str, unicode)):
                    if value.startswith('/'):
                        pass
                    else: value = os.path.join(self.root, value)
                elif hasattr(value, '__iter__'):
                    if hasattr(value, 'items'):
                        value = dictWrapper(value, self.root)
                    else:
                        value = listWrapper(value, self.root)
            if value is _Null:
                raise AttributeError('Attribute %r not found' % name)
            return value


    __yaml_keys = collections.defaultdict(list)
    __yaml_files = collections.defaultdict(list)
    __options = collections.defaultdict(set)
    __all__ = []

    @property
    def EMEN2DBHOME(self):
        '''The root for relative paths, also the base directory of the DB'''
        return self.getattr('EMEN2DBHOME')

    @EMEN2DBHOME.setter
    def EMEN2DBHOME(self, value):
        # Whenever EMEN2DBHOME is set, it needs to update paths.root.
        self.setattr('EMEN2DBHOME', value)
        self.paths.root = value

    @classmethod
    def check_locked(cls):
        try:
            cls.__locked
            #print 'ns __locked 0', cls.__locked
        except AttributeError:
            cls.__locked = False
        return cls.__locked
        #print 'ns __locked 1', cls.__locked

    @property
    def _locked(self):
        return self.check_locked()

    @_locked.setter
    def _locked(self, value):
        cls = self.__class__
        if not self.__locked:
            cls.__locked = bool(value)
        elif bool(value) == False:
            raise LockedNamespaceError, 'cannot unlock %s' % cls.__name__


    @classmethod
    def __unlock(cls):
        """unlock namespace (for internal/debug use only)"""
        cls.__locked = False


    hier = dict()

    @classmethod
    def load_data(cls, fn, data):
        '''Load data from a specified configuration file (either a YAML file or a JSON one) or dictionary

        :attr fn: the name of the configuration file
        :attr data: a dictionary to load the data from
        '''
        fn = os.path.abspath(fn)
        if fn and os.access(fn, os.F_OK):
            ext = fn.rpartition('.')[2]
            with open(fn, "r") as f:
                data = f.read()

            loadfunc = {'json': json.loads}
            try:
                loadfunc['yml'] = yaml.safe_load
            except AttributeError:
                pass

            def fail(*a):
                raise NotImplementedError, "No loader for %s files found" % ext.upper()

            loadfunc = loadfunc.get( fn.rpartition('.')[2], fail )

            if ext.lower() == 'json':
                data = json_strip_comments(data)

            data = loadfunc(data)

        elif hasattr(data, 'upper'):
            loadfunc = json.loads
            if yaml:
                load_func = yaml.safe_load
            data = loadfunc(data)

        return data


    @classmethod
    def from_file(cls, fn=None, data=None):
        """Alternate constructor which initializes a GlobalNamespace instance from a JSON or YAML file"""

        if not (fn or data):
            raise ValueError, 'Either a filename or json/yaml data must be supplied'

        data = cls.load_data(fn, data)

        ## load data
        self = cls()

        if data:
            # emen2.db.log.info("Loading config: %s"%fn)

            # treat EMEN2DBHOME specially
            self.EMEN2DBHOME = data.pop('EMEN2DBHOME', self.getattr('EMEN2DBHOME', ''))
            self.paths.root = self.EMEN2DBHOME

            for k,v in data.pop('paths', {}).items():
                setattr(self.paths, k, v)

            # process data
            self._create = True
            for key in data:
                b = data[key]
                pref = ''.join(b.pop('prefix',[])) # get the prefix for the current toplevel dictionary
                options = b.pop('options', {})      # get options for the dictionary
                self.__yaml_files[fn].append(key)

                for key2, value in b.iteritems():
                    self.__yaml_keys[key].append(key2)
                    #self.setattr('.'.join([key,key2]), value, options)
            self.from_dict(data)

            # load alternate config files
            # for fn in self.paths.configfiles:
            #    fn = os.path.abspath(fn)
            #    if os.path.exists(fn):
            #        cls.from_file(fn=fn)

        return self


    def to_json(self, keys=None, kg=None, file=None, indent=4, sort_keys=True):
        '''Dump the data as a JSON file'''
        return json.dumps(self.__dump_prep(keys, kg, file), indent=indent, sort_keys=sort_keys)


    def to_yaml(self, keys=None, kg=None, file=None, fs=0):
        '''Dump the data as a YAML file'''
        return yaml.safe_dump(self.__dump_prep(keys, kg, file), default_flow_style=fs)

    def __dump_prep(self, keys=None, kg=None, file=None):
        """store state as YAML"""
        if keys is not None:
            keys = keys
        elif kg is not None:
            keys = dict( (k, self.__yaml_keys[k]) for k in kg)
        elif file is not None:
            keys = dict( (k, self.__yaml_keys[k]) for k in self.__yaml_files[file] )
        else:
            keys = self.__yaml_keys

        _dict = collections.defaultdict(dict)
        for key, value in keys.iteritems():
            for k2 in value:
                value = self.getattr(key).getattr(k2)
                if hasattr(value, 'json_equivalent'): value = value.json_equivalent()
                _dict[key][k2] = value

        for key in self.paths.attrs:
            path = getattr(self.paths, key)
            if hasattr(path, 'json_equivalent'): path = path.json_equivalent()
            _dict['paths'][key] = path
        return dict(_dict)

    def json_equivalent(self):
        '''Return a JSON equivalent of this class, for :py:func:`jsonutil.encode`'''
        return dict(self._values)

    #@classmethod
    #def setattr(self, name, value):
    #    self.__addattr(name, value)

    #def __setattr__(self, name, value):
    #    self._trigger_event(name, value, write=True)
    #    res = getattr(self.__class__, name, None)
    #    if name.startswith('_') or hasattr(res, '__set__'):
    #        object.__setattr__(self, name, value)
    #    else:
    #        self.setattr(name, value)

    def __getitem__(self, name):
        if not hasattr(name, 'split'):
            name = '.'.join(name)
        return self.getattr(name)

    def __setitem__(self, name, value):
        if not hasattr(name, 'split'):
            name = '.'.join(name)
        return self.setattr(name, value)

    #@classmethod
    def setattr(self, name, value, options=None):
        '''Set an attribute'''
        if not name in self.__all__:
            self.__all__.append(name)

        self.check_locked()
        if name.startswith('_') or not self.__locked:
            if options is not None:
                for option in options:
                    self.__options[option].add(name)
            Hier.setattr(self, name, value)
        else: raise LockedNamespaceError, 'cannot change locked namespace'

    def getattr(self, name, default=None, *args, **kwargs):
        '''Get an attribute, returning default if not accessible'''
        if name.endswith('___'):
            name = name.partition('___')[-1]
        if self.__options.has_key('private'):
            if name in self.__options['private']:
                return default
        result = Hier.getattr(self, name, default=default, *args, **kwargs)
        return result





#import <module>
import unittest

class TestGlobalNamespace(unittest.TestCase):

    def setUp(self):
        self.a = GlobalNamespace('one instance')
        self.b = GlobalNamespace('two instance')
        self.a.a = 1


    def test_attributenotfound(self):
        self.assertRaises(lambda: self.a.not_found, AttributeError)

    def test_setattribute(self):
        #test 1 attribute access
        self.assertEqual(self.a.a, 1)
        self.assertTrue(hasattr(self.b, 'a'))
        self.assertEqual(self.a.a, self.b.a)

    def test_reset(self):
        self.a.reset()
        self.assertRaises(lambda: self.a.a, AttributeError)

    def test_update(self):
        tempdict = dict(a=1, b=2, c=3)
        self.a.update(tempdict)
        self.assertEqual(self.a.a, tempdict['a'])
        self.assertEqual(self.b.a, tempdict['a'])
        self.assertEqual(self.a.___a, tempdict['a'])
        print "test 3 passed"
        a.reset()


if __name__ == '__main__':
    pass
    #unittest.main()

class dictWrapper(object, UserDict.DictMixin):
    def __init__(self, dict_, prefix):
        self.__dict = dict_
        self.__prefix = prefix
    def __repr__(self):
        return '<dictWrapper dict: %r>' % self.__dict
    def keys(self): return self.__dict.keys()
    def __getitem__(self, name):
        v = self.__dict[name]
        if not v.startswith('/'):
            v = os.path.join(self.__prefix, v)
        return v
    def __setitem__(self, name, value):
        if isinstance(value, (str, unicode)):
            if value.startswith(self.__prefix):
                del value[len(self.__prefix)+1:]
        self.__dict[name] = value
    def __delitem__(self, name):
        del self.__dict[name]
    def json_equivalent(self): return dict(self)

class listWrapper(object):
    def __init__(self, list_, prefix):
        self.__list = list_
        self.__prefix = prefix
    def __repr__(self):
        return '<listWrapper list: %r, prefix: %s>'%(self.__list, self.__prefix)
    def check_item(self, item):
        if isinstance(item, (str, unicode)):
            item = os.path.join(self.__prefix, item)
        return item
    def __iter__(self):
        for item in self.__list:
            yield self.check_item(item)
    def __getitem__(self, k):
        return self.check_item(self.__list[k])
    def chopitem(self, item):
        if isinstance(item, (str, unicode)):
            if item.startswith(self.__prefix):
                item = item[len(self.__prefix)+1:]
        return item
    def __setitem__(self, key, value):
        value = self.chopitem(value)
        self.__list[key] = value
    def __delitem__(self, idx):
        del self.__list[idx]
    def append(self, item):
        item = self.chopitem(item)
        self.__list.append(item)
    def extend(self, items):
        self.__list.extend(self.chopitem(item) for item in items)
    def pop(self, idx):
        res = self[idx]
        del self[idx]
        return res
    def count(self, item):
        return self.__list.count(self.chopitem(item))
    def insert(self, idx, item):
        self.__list.insert(idx, self.chopitem(item))
    def __len__(self): return len(self.__list)
    def json_equivalent(self): return list(self)



class Watch(object):
    def __init__(self, ns, name, default, validator=None):
        self.ns = ns
        self.name = name.split('.')
        self.default = default

    def __get__(self, instance, owner):
        return self.get()

    def get(self):
        result = getattr(self.ns, self.name[0], self.default)
        for n in self.name[1:]:
            result = getattr(result, n, self.default)

        return result

class Claim(Watch):
    claimed_attributes = set()
    def __init__(self, ns, name, default, validator=None):
        if name in self.claimed_attributes:
            raise ValueError, 'attribute %s already claimed, use GlobalNamespace.watch() instead' % name
        else:
            self.claimed_attributes.add(name)

        if not hasattr(ns, name):
            ns.setattr(name, default),

        self.validator = validator
        self._validate(default)


        Watch.__init__(self, ns, name, default, validator)

    def _validate(self, value):
        if self.validator is not None:
                is_valid = self.validator(value)
                if not is_valid:
                    res_repr = repr(value)
                    if len(res_repr) > 30: res_repr = res_repr[:27] + '...'
                    raise ValueError, 'Configuration value %r has invalid value %s' % (self.name, res_repr)

    def __set__(self, instance, value):
        self.set(value)

    def set(self, value):
        self._validate(value)

        ns = self.ns
        for name in self.name[:-1]:
            ns = getattr(ns, name)

        setattr(ns, self.name[-1], value)
        return self



def json_strip_comments(data):
    r = re.compile('/\\*.*\\*/', flags=re.M|re.S)
    data = r.sub("", data)
    data = re.sub("\s//.*\n", "", data)
    return data

class LockedNamespaceError(TypeError):
    pass


# try:
#     import emen2.db
#     emen2.db.log
# except AttributeError:
#     import emen2.db.log

__version__ = "$Revision: 1.100 $".split(":")[1][:-1].strip()


# $Id: util.py,v 1.1 2012/10/19 08:19:34 irees Exp $

# Temporary fix --
import jsonrpc.jsonutil
def jsonencode(*args, **kwargs):
    return jsonrpc.jsonutil.encode(*args, **kwargs).replace('/', r'\/')


# $Id: load.py,v 1.42 2013/05/13 23:33:09 irees Exp $
"""Load a dumped database

Functions:
    random_password

Classes:
    BaseLoader
    Loader
"""

import os
import sys
import time
import tarfile
import tempfile
import string
import random
import collections
import getpass
import json
import jsonrpc.jsonutil

# EMEN2 imports
import emen2.util.listops
import emen2.db.config


def random_password(N):
    """Generate a random password of length N."""
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for x in range(N))

class BaseLoader(object):
    """Load database objects from a JSON file."""
    def __init__(self, db=None, infile=None):
        self.infile = infile
        self.db = db

    def load(self):
        pass

    def loadfile(self, infile=None, keytype=None):
        infile = infile or self.infile
        if not os.path.exists(infile):
            yield
        with open(infile) as f:
            for item in f:
                item = item.strip()
                if item and not item.startswith('/'):
                    item = jsonrpc.jsonutil.decode(item)
                    if keytype:
                        if keytype == item.get('keytype'):
                            yield item
                    else:
                        yield item


class Loader(BaseLoader):
    # Speeds:

    # Single txn:
    #   Reading: 25000/s
    #   Instantiating: 10500/s
    #   Loading/validating: 1900/s
    #   Writing: 200/s

    # Separate txns:
    #   Reading file: 25000/s
    #   Opening txn: 8500/s
    #   Instantiating: 6000/s
    #   Loading/validating: 1600/s
    #   Writing: 83/s

    def load(self):
        t = time.time()
        count = 0
        dbenv = self.db._db.dbenv
        ctx = self.db._ctx
        txn = self.db._txn
        for item in self.loadfile():
            print "\n======= put count %s"%count
            keytype = item.get('keytype')
            i = dbenv[keytype].dataclass(ctx=ctx) 
            i._load(item)
            dbenv[keytype]._put(i, ctx=ctx, txn=txn)
            count += 1
        t = time.time()-t
        s = float(count) / t
        print "total time: %s, %s put/sec"%(t, s)


class LoadOptions(emen2.db.config.DBOptions):
    def parseArgs(self, infile):
        self['infile'] = infile

if __name__ == "__main__":
    import emen2.db
    cmd, db = emen2.db.opendbwithopts(optclass=LoadOptions, admin=True)
    with db._newtxn(write=True):
        loader = Loader(db=db, infile=cmd.options['infile'])
        loader.load()
            



# $Id: queues.py,v 1.12 2012/07/31 21:46:10 irees Exp $
"""EMEN2 process and message queues
"""

import time
import random
import subprocess
import threading
import Queue


# try:
#    import multiprocessing
#    CPU_COUNT = multiprocessing.cpu_count()
# except:
#    CPU_COUNT = 1

CPU_COUNT = 2

import emen2.db.log

##### Subprocess queues #####

# threading.Thread style
class ProcessWorker(object):
    def __init__(self, queue):
        self.queue = queue

    def run(self):
        while True:
            priority, name, task = self.queue.get()
            if task is None:
                self.queue.add_task(None)
                return
            desc = ' '.join(map(str, task))
            emen2.db.log.info("ProcessWorker run: %s"%(desc))
            # a = subprocess.Popen(task)
            # returncode = a.wait()
            try:
                ret = subprocess.check_output(task, stderr=subprocess.STDOUT)
            except Exception, e:
                emen2.db.log.error("Couldn't build tile: %s"%e)
            finally:
                self.queue.task_done()
            
            

class ProcessQueue(Queue.LifoQueue):
    """A queue of processes to run."""
    worker = ProcessWorker
    
    def add_task(self, task, priority=0, name=None):
        """Add a task to the queue. Highest priority is 0, lowest priority is 1000. Default is 100."""
        if task:
            task = tuple(task)

        if priority < -100 or priority > 100:
            raise ValueError, "Highest priority is -100, lowest priority is 100"

        if name in [i[1] for i in self.queue]:
		priority = 0
        #    raise ValueError, "Task name already in queue: %s"%name

        emen2.db.log.info("ProcessQueue: Adding task %s with priority %s to queue: %s"%(name, priority, task))
        return self.put((priority, name, task))

    def end(self):
        """Allow all queued jobs to complete, then exit."""
        return self.put((101, None))

    def stop(self):
        """Allow running jobs to complete, then exit."""
        return self.put((-101, None))

    def start(self, processes=1):
        for i in range(processes):
            worker = self.worker(self)
            t = threading.Thread(target=worker.run)
            t.daemon = True
            t.start()
            

processqueue = ProcessQueue()
processqueue.start(processes=CPU_COUNT)


if __name__ == "__main__":
    pq = ProcessQueue()
    for i in range(20):
        pq.add_task(['touch', '%s.txt'%i])
    pq.start()
    pq.join()
    # pq.end()


__version__ = "$Revision: 1.12 $".split(":")[1][:-1].strip()



# $Id: view.py,v 1.146 2013/01/04 20:52:22 irees Exp $
import sys
import os
import time
import collections
import functools

import jsonrpc.jsonutil

# emen2 imports
import emen2.web.routing
import emen2.web.resource
import emen2.db.config
import emen2.db.log


##### I. Views #####

class TemplateContext(collections.MutableMapping):
    '''Template context.
    
    This is a dict-like object that will be passed into the template renderer.
    '''

    def __init__(self, d=None):
        self.__dict = {}
        self.__dict['ctxt'] = self
        self.__dict.update(d or {})
        
        self.notify = []
        self.errors = []
        
        # Page title and template
        self.title = 'No title'
        self.template = '/simple'

        # Configuration settings
        self.root = emen2.db.config.get('web.root')
        self['TITLE'] = emen2.db.config.get('customization.title')
        self.version = emen2.__version__


    def __getitem__(self, n):
        return self.__dict[n]

    def __setitem__(self, n, v):
        self.__dict[n] = v

    def __delitem__(self, n):
        del self.__dict[n]

    def __len__(self):
        return len(self.__dict)

    def __iter__(self):
        return iter(self.__dict)

    def __repr__(self):
        return '<TemplateContext: %r>'%self.__dict

    def copy(self):
        return TemplateContext(self.__dict)

    def set(self, name, value=None):
        self[name] = value

    def reverse(self, _name, *args, **kwargs):
        """Create a URL given a view Name and arguments"""
        
        host = emen2.db.config.get('web.host')
        port = emen2.db.config.get('web.port')
        full = kwargs.pop('_full', False)

        result = emen2.web.routing.reverse(_name, *args, **kwargs)
        result = result.replace('//','/')
        if full:
            result = 'http://%s:%s%s' % (host, port, result)

        containsqs = '?' in result
        if not result.endswith('/') and not containsqs:
            result = '%s/' % result
        elif containsqs and '/?' not in result:
            result = result.replace('?', '/?', 1)

        return result



class TemplateView(emen2.web.resource.EMEN2Resource):
    '''An EMEN2Resource class that renders a result using a template.'''

    # Basic properties
    title = property(
        lambda self: self.ctxt.title,
        lambda self, value: setattr(self.ctxt, "title", value))

    template = property(
        lambda self: self.ctxt.template,
        lambda self, value: setattr(self.ctxt, "template", value))

    def init(self, *arrgghs, **blarrgghs):
        # Template context
        self.ctxt = TemplateContext()
        
        # ETags
        self.etag = None

    #### Output methods #####

    def notify(self, msg, error=False):
        if not msg:
            return
        if error:
            self.ctxt.errors.append(msg)            
        else:
            self.ctxt.notify.append(msg)

    def simple(self, title=None, content=None):
        '''Set the output to a simple message.'''
        self.template = '/simple'
        self.title = title
        self.ctxt['content'] = content

    def error(self, msg):
        '''Set the output to a simple error message.'''
        self.template = "/errors/error"
        self.title = 'Error'
        self.ctxt['errmsg'] = msg

    def redirect(self, redirect=None, title=None, content='', auto=True, showlink=True):
        '''Redirect by setting Location header and using the redirect template.'''
        content = content or ''
        self.template = '/redirect'
        self.title = title or self.title
        self.ctxt['content'] = content        
        self.ctxt['showlink'] = showlink
        redirect = redirect or '%s/'%emen2.db.config.get('web.root')
        self.ctxt['redirect'] = redirect
        if auto:
            self._redirect = redirect.replace('//','/')

    def get_data(self):
        '''Render the template.'''
        return emen2.db.config.templates.render_template(self.ctxt.template, self.ctxt)




class View(TemplateView):
    '''A View with a database connection.'''

    def init(self, *args, **kwargs):
        '''Run this before the requested view method.'''
        super(View, self).init(*args, **kwargs)
        user = {}
        admin = False
        ctx = getattr(self.db, '_getctx', lambda:None)()
        try:
            user = ctx.db.user.get(ctx.username)
            admin = ctx.checkadmin()
        except:
            pass

        self.ctxt.update(dict(
            USER = user,
            ADMIN = admin,
            DB = self.db
        ))


__version__ = "$Revision: 1.146 $".split(":")[1][:-1].strip()



# $Id: tiles.py,v 1.15 2012/10/18 23:34:23 irees Exp $
import os
import pickle
import math
import json

import jsonrpc.jsonutil

import emen2.db.config
from emen2.web.view import View

import emen2.db.handlers

# header[index][slices][tiles][(level, x, y)]


@View.register
class Preview(View):
    @View.add_matcher(r'^/preview/(?P<bid>[^/]*)/(?P<mode>.+)/$')    
    def main(self, bid=None, mode='tiles', **kwargs):
        if bid == None:
            return "No Binary ID supplied."

        # Make sure we can access bdo
        bdo = self.db.binary.get(bid, filt=False)
        filename = bdo.get('filename')
        size = int(kwargs.get('size', 512))
        index = int(kwargs.get('index', 0))
        scale = int(kwargs.get('scale', 1))
        z = int(kwargs.get('z', 0))
        x = int(kwargs.get('x', 0))
        y = int(kwargs.get('y', 0))

        # get_data(self)
        previewpath = emen2.db.binary.Binary.parse(bid).get('previewpath')
        previewpath = '%s.eman2'%(previewpath)

        if not os.path.exists(previewpath):
            status = emen2.db.handlers.thumbnail_from_binary(bdo, wait=False)
            raise Exception, "Building tile..."

        f = file(previewpath, "r")
        header = pickle.load(f)

        if mode == 'header':
            h = header[index]
            data = {
                'nx': h['nx'],
                'ny': h['ny'],
                'nz': h['nz'],
                'maxscale': 8,
                'filename': filename
            }
            f.close()
            return jsonrpc.jsonutil.encode(data)

        h = header[index]['slices'][z]
        key = size
        if mode == 'tiles':
            key = (scale, x, y)

        ret = h[mode][key]

        f.seek(ret[0], 1)
        data = f.read(ret[1])
        f.close()

        if ret[2] == 'jpg':
            self.set_header("Content-Type", "image/jpeg")
        elif ret[2] == 'png':
            self.set_header("Content-Type", "image/png")
        elif ret[2] == 'json':
            self.set_header("Content-Type", "application/json")

        return data
    




__version__ = "$Revision: 1.15 $".split(":")[1][:-1].strip()



# $Id: log.py,v 1.28 2013/05/08 04:51:38 irees Exp $
"""EMEN2 logging

Functions:
    msg
    info
    init
    error
    warn
    debug
    security
    index
    commit
"""


import traceback
import functools
import os.path
import sys

import twisted.python.log

from twisted.python import util
from twisted.python.log import _safeFormat, textFromEventDict

import emen2.db.config

class PrintLogger(object):
    def emit(self, eventDict):
        print eventDict

class SubLogger(twisted.python.log.FileLogObserver):
    pass

class WebLogger(twisted.python.log.FileLogObserver):
    def emit(self, eventDict):
        messages = eventDict.get("message")
        for message in messages:
            util.untilConcludes(self.write, message+"\n")
            util.untilConcludes(self.flush)

class ErrorLogger(twisted.python.log.FileLogObserver):
    pass
    
class EMEN2Logger(object):
    log_levels = dict(
            DEBUG=-1,
            TXN=1,
            INIT=2,
            INDEX=3,
            INFO=4,
            COMMIT=5,
            WEB=6,
            WARN=8,
            SECURITY=9,
            ERROR=10,
            CRITICAL=11,
            ALL=0
    )
    
    def __init__(self):
        """Initialize logging system."""
        # print "EMEN2Logger.__init__"
        self.started = False
        self.log_level = 0
        self.loggers = {}
        # Turn on logging to stdout by default
        # twisted.python.log.startLogging(sys.stdout, setStdout=False)

    def init(self):
        """Start logging system."""
        # The configuration has been loaded
        self.logpath = emen2.db.config.get("paths.log")
        self.log_level = self.log_levels.get(emen2.db.config.get('log_level', 0))

    def start(self):
        """Start file-backed logging."""
        self.started = True

        # Open the various log files.        
        if not self.logpath:
            raise Exception, "No log path set"
            
        self.loggers["INFO"] = SubLogger(open(os.path.join(self.logpath, "emen2.log"), "a+"))
        self.loggers["SECURITY"] = SubLogger(open(os.path.join(self.logpath, "security.log"), "a+"))
        self.loggers["ERROR"] = ErrorLogger(open(os.path.join(self.logpath, "error.log"), "a+"))
        self.loggers["WEB"] = WebLogger(open(os.path.join(self.logpath, "access.log"), "a+"))

    def stop(self):
        """Stop file-backed logging."""
        self.started = False
        for k,v in self.loggers.items():
            v.close()

    def emit(self, e):
        """Twisted log file observer function."""
        level = e.get("system", "INFO")
        output = self.loggers.get(level, self.loggers["INFO"])
        output.emit(e)

    def log(self, message, level='INFO'):
        """Print or write the log message."""
        priority = self.log_levels.get(level, 0)        
        if priority < self.log_level:
            return

        # If we're using twisted logging, pass through...
        if self.started: 
            twisted.python.log.msg(message, system=level)
        else:
            pass
            print "[%s]"%level, unicode(message).encode('utf-8')

# Create the logger
logger = EMEN2Logger()

def print_exception():
    traceback.print_exc()

def msg(msg='', level='INFO'):
    logger.log(msg, level)

# Aliases
info = functools.partial(msg, level='INFO')
init = functools.partial(msg, level='INIT')
error = functools.partial(msg, level='ERROR')
warn = functools.partial(msg, level='WARN')
debug = functools.partial(msg, level='DEBUG')
security = functools.partial(msg, level='SECURITY')
index = functools.partial(msg, level='INDEX')
commit = functools.partial(msg, level='COMMIT')
web = functools.partial(msg, level='WEB')



import collections
import os
import json

import jsonrpc.jsonutil
import emen2.db
import emen2.db.config
import emen2.util.listops


def dump_json(outfile, items):
    if os.path.exists(outfile):
        print "Warning: File %s exists"%outfile
    print "Saving output to %s"%outfile
    with open(outfile,'w') as f:
        for item in items:
            f.write(json.dumps(item)+"\n")


class Dumper(object):
    
    keytypes = ['paramdef', 'recorddef', 'user', 'group', 'binary', 'record']
    
    def __init__(self, db):
        self.db = db
        self._ref_pds = {}
        self._ref_vts = collections.defaultdict(set)
        
    def dump(self, names=None, keytype='record', **kwargs):
        # Query
        kwargs['keytype'] = keytype
        names = names or self.db.query(**kwargs)['names']
        print "Starting with:"
        print len(names)
        keys = collections.defaultdict(set)
        keys[keytype] |= set(names)

        # Find referenced items.
        for chunk in emen2.util.listops.chunk(keys[keytype], count=1000):
            items = self.db.get(chunk, keytype=keytype)
            print "Processing: %s ... %s"%(items[0].name, items[-1].name)
            # Find paramdefs that reference other items..
            self._findrefs(items)
            # Find other items..
            for kt in self.keytypes:
                keys[kt] |= getattr(self, '_find_%s'%kt)(items)

        return keys
        
    def write(self, keys, outfile='dump.json', uri=None):
        f = open(outfile, 'w')
        for keytype in self.keytypes:
            modify = getattr(self, '_modify_%s'%keytype, lambda x:x)
            
            for chunk in emen2.util.listops.chunk(keys.get(keytype, []), count=1000):
                items = self.db.get(chunk, keytype=keytype)
                for item in items:
                    if uri:
                        item.__dict__['uri'] = '%s/%s/%s'%(uri, keytype, item.name)
                    item = modify(item)
                    print "%s: %s"%(item.keytype, item.name)
                    f.write(jsonrpc.jsonutil.encode(item))
                    f.write('\n')
        f.close()
        return outfile        

    def _findrefs(self, items):
        # Get the referencing vartypes from the items.
        pds = self._find_paramdef(items)
        for param in pds - set(self._ref_pds.keys()):
            pd = self.db.get(param, keytype='paramdef')
            if pd.vartype in self.keytypes:
                self._ref_pds[pd.name] = pd.vartype
                self._ref_vts[pd.vartype].add(pd.name)
    
    def _findvalues(self, vartypes, items):
        # Get the values referenced by a vartype from the items.
        ret = set()
        pds = set()
        for vartype in vartypes:
            pds |= self._ref_vts[vartype]
        for pd in pds:
            pd = self.db.paramdef.get(pd)
            if pd.iter:
                for item in items:
                    ret |= set(item.get(pd.name, []))
            else:
                for item in items:
                    ret.add(item.get(pd.name))
        ret -= set([None])
        return ret

    def _find_paramdef(self, items):
        r = set()
        for rec in items:
            r |= set(rec.keys())
        return r

    def _find_record(self, items):
        return self._findvalues(['record'], items)
            
    def _find_recorddef(self, items):
        return self._findvalues(['recorddef'], items)
        
    def _find_user(self, items):
        return self._findvalues(['user'], items)
                
    def _find_group(self, items):
        return self._findvalues(['group'], items)

    def _find_binary(self, items):
        return self._findvalues(['binary'], items)

    def _modify_record(self, item):
        return item
        
    def _modify_user(self, item):
        return item



class PublicDumper(Dumper):
    def _modify_record(self, item):
        # Remove permissions, mark published data as anon
        if set(['authenticated', 'publish']) & item.__dict__['groups']:
            item.__dict__['groups'] = set(['anon'])
        else:
            item.__dict__['groups'] = set()
        item.__dict__['permissions'] = [[],[],[],[]]
        return item
        
    def _modify_user(self, item):
        # Remove user passwords
        item.__dict__['password'] = ''
        return item



class DumpOptions(emen2.db.config.DBOptions):
    def parseArgs(self, infile):
        self['infile'] = infile



if __name__ == "__main__":
    import emen2.db
    db = emen2.db.opendb(admin=True)
    dumper = Dumper(db=db) # PublicDumper(db=db)
    keys = dumper.dump(c=[['groups','==','publish']])
    keys['paramdef'] = db.paramdef.filter(None)
    keys['recorddef'] = db.recorddef.filter(None)
    keys['user'] = db.user.filter()
    keys['user'].remove('root')
    dumper.write(keys, outfile="dump-publicdata.json", uri="http://ncmidb.bcm.edu")
    
    
    



# $Id: create.py,v 1.2 2013/05/14 15:33:23 irees Exp $
"""Create a new database."""

import os
import sys
import time
import tarfile
import tempfile
import string
import random
import collections
import getpass
import json
import jsonrpc.jsonutil

# EMEN2 imports
import emen2.util.listops
import emen2.db.config



class CreateOptions(emen2.db.config.DBOptions):
    def parseArgs(self):
        pass

if __name__ == "__main__":
    import emen2.db
    db = emen2.db.opendb(admin=True)
    db._db.dbenv.create()
    emen2.db.database.setup(db=db)
            
        


# $Id: handlers.py,v 1.39 2013/02/28 00:49:55 irees Exp $
'''File handlers

'''
import shutil
import time
import re
import collections
import traceback
import math
import os
import signal
import optparse
import subprocess
import sys

# For file writing
import shutil
import hashlib
import cStringIO
import tempfile

##########################################################
# ** Do NOT import ANY EMEN2 or EMAN2 packages here!! ** #
##########################################################

def thumbnail_from_binary(bdo, force=False, wait=False, priority=0):
    """Given a Binary instance, run the thumbnail builder as a separate process
    
    Returns a status:
        "completed"
        "building"
        "error"

    """
    # Import the EMEN2 modules here to prevent circular imports.
    import emen2.db.config
    import emen2.db.binary
    import emen2.db.queues
    
    # Paths and filenames
    previewpath = emen2.db.binary.Binary.parse(bdo.get('name')).get('previewpath')
    filepath = bdo.get('filepath')
    filename = bdo.get('filename')

    # Sanitize the filename to check compress= and ext=
    ext = ''
    compress = ''
    r = re.compile('[\w\-\.]', re.UNICODE)
    _filename = "".join(r.findall(filename)).lower()
    _filename = _filename.split(".")
    if _filename and _filename[-1] in ['gz', 'bz2']:
        compress = _filename.pop()
    if _filename:
        ext = _filename.pop()

    # Get the handler.
    handler = BinaryHandler.get_handler(filepath=filepath, filename=filename)

    # If the thumbnail is currently building, or an error occurred previously,
    # then return that status.
    status = handler.thumbnail_status(previewpath)
    if status in ["building", "error"]:
        return status
    
    # Prepare the command to run.
    
    # Grumble... Come up with a better way to get the script name.
    args = [sys.executable]
    cmd = emen2.db.config.get_filename(handler.__module__)
    fix = ['.pyc', '.pyo']
    if cmd[-4:] in fix:
        for f in fix:
            cmd = cmd.replace(f, '.py')

    args.append(cmd)

    # Add the command arguments.
    args.append('--previewpath')
    args.append(previewpath)
    
    if compress:
        args.append('--compress')
        args.append(compress)

    args.append(handler.__class__.__name__)
    args.append(filepath)
    
    if wait:
        # Run directly and wait.
        a = subprocess.Popen(args)
        a.wait()
        return "complete"
    
    # Otherwise, add to the task queue.
    emen2.db.queues.processqueue.add_task(args, name=filepath, priority=priority)
    return "building"
    
    

def main(g):
    """Use a Handler to build a thumbnail."""
    # g is the module namespace...
    parser = ThumbnailOptions()
    options, (handler, filepath) = parser.parse_args()
    
    filename = 'filename'
    if options.ext:
        filename = '%s.%s'%(filename, options.ext)
    if options.compress:
        filename = '%s.%s'%(filename, options.compress)
    
    # Get the handler and build the thumbnail.
    handler = g[handler](filepath=filepath, filename=filename)
    handler.thumbnail(previewpath=options.previewpath, force=options.force)
    
    
    
##### Exceptions and Options #####    

class ThumbnailError(Exception):
    pass


class ThumbnailOptions(optparse.OptionParser):
    """Options to run a Binary thumbnail generator."""
    
    def __init__(self, *args, **kwargs):
        optparse.OptionParser.__init__(self, *args, **kwargs)
        self.add_option('--previewpath', type="string", help="Base output")
        self.add_option('--ext', type="string", help="Original filename extension")
        self.add_option('--compress', type="string", help="Compression type")
        self.add_option('--force', action="store_true", help="Force rebuild")
        usage = """%prog [options] <handler class> <input file>"""

    def parse_args(self, lc=True, *args, **kwargs):
        options, args = optparse.OptionParser.parse_args(self,  *args, **kwargs)
        if len(args) < 2:
            raise ValueError, "Handler class name and input file required"
        return options, args



##### File handler #####

class BinaryHandler(object):
    '''EMEN2 managed file.

    This is used in a few different ways. It is used for files uploaded
    to the web server. It can also be used with the File Handlers defined
    below (e.g. to extract header metadata.) Handlers can also be passed
    as items to db.binary.put.
    
    The original name of the file is self.filename. Sources can be a added in
    the constructor, and may be a string of data (filedata), or a file-like
    object supporting read() (fileobj). Consider the data to be read-only.
    '''

    # The self._getfilepath() method will return an on-disk filename that can be used
    # for operations that required a named file (e.g. EMAN2.) If the input
    # source is filedata or fileobj, it will write out to a temporary file in
    # the normal temp file storage area. The close() method will remove any
    # temporary files.
    
    # File type handlers
    _handlers = {}
    _allow_gzip = False

    def __init__(self, filename, filedata=None, fileobj=None, filepath=None, param='file_binary'):
        # Original filename
        self.filename = filename
        if not self.filename:
            raise ThumbnailError, "Filename required"

        # Parameter in the associated record
        self.param = param
        
        # Temporary files
        self._tmpfiles = []

        # One of the following is required: filepath / filedata / fileobj
        # ian: note... doesn't have to be set during init -- the 
        # web resource will do it later for PUT/POST'd files.
        self.filepath = filepath
        self.filedata = filedata
        self.fileobj = fileobj
        # if not any([self.filepath, self.filedata, self.fileobj]):
        #    raise ThumbnailError, "No data; can be filepath, filedata, or fileobj."


    def get(self, key, default=None):
        return self.__dict__.get(key, default)


    ##### Open the underlying data #####
        
    def open(self):
        '''Open the file.
        :return: File-like object
        '''
        
        if self.filepath:
            # If there is a filepath, open and return that
            return open(self.filepath, "r")

        readfile = None
        if self.filedata:
            # Take a StringIO or string and make a StringIO
            readfile = cStringIO.StringIO(self.filedata)
        elif self.fileobj:
            # ... use the fileobj
            self.fileobj.seek(0)
            readfile = self.fileobj
        else:
            raise IOError, "No file given, or don't know how to read file.."
        return readfile

    # Making this a private method for now
    def _getfilepath(self, suffix=None):
        '''Write to temporary storage.
        :return: Temporary file path.
        '''
        
        if self.filepath:
            # If there is a filepath, return that
            return self.filepath
        
        # .. or copy the filedata/fileobj to a temporary file
        infile = self.open()

        # Make a temporary file
        suffix = suffix or '.tmp'
        (fd, tmpfile) = tempfile.mkstemp(suffix=suffix)
        with os.fdopen(fd, "w+b") as f:
            shutil.copyfileobj(infile, f)

        # infile.close()
        # TODO: Set as self.filepath, for subsequent calls to self._getfilepath().
        # TODO: Should probably fine a better way to remove the tmpfile.
        self._tmpfiles.append(tmpfile)
        return tmpfile

    def close(self):
        # Remove temporary files...
        for f in self._tmpfiles:
             os.remove(f)


    ##### Extract metadata and build thumbnails #####

    def extract(self, **kwargs):
        return {}

    def thumbnail(self, previewpath, force=False):
        fp = self._getfilepath()
        if not os.access(fp, os.F_OK):
            raise ThumbnailError, "Could not access: %s"%fp

        # This is used in self._outfile()
        self._previewpath = previewpath

        # Make sure the output directory exists..
        pdir = os.path.dirname(self._previewpath)
        if not os.path.exists(pdir):
            try:
                os.makedirs(pdir)
            except OSError:
                pass

        # Create a status file to indicate work on this thumbnail has started
        statusfile = self._outfile('status')
        
        # Write out "building" to the status file.
        with file(statusfile, 'w') as f:
            f.write("building")

        # Check compression and file type
        compress = False
        ext = ''
        if self.filename:
            fn = self.filename.split(".")
            if fn[-1] == 'gz':
                compress = 'gzip'
                fn.pop()
            ext = fn.pop()

        # Build the thumbnail.
        if compress and not self._allow_gzip:
            # This handler does not accept gzip'd files, for whatever reason.
            pass
            
        elif compress:
            # Decompress the file
            workfile = tempfile.mkstemp(suffix='.tmp')[1]
            cmd = "%s -d -c %s > %s"%(compress, fp, workfile)
            os.system(cmd)
            try:
                self._thumbnail_build(workfile)
            except Exception, e:
                emen2.db.log.error("Could not build tiles: %s"%e)
                pass
            os.remove(workfile)
        else:
            self._thumbnail_build(fp)

        # Finished the thumbnail; remove the status file.
        try:
            os.remove(statusfile)
        except:
            pass

    def _thumbnail_build(self, workfile, **kwargs):
        # Override this method to actually build the thumbnails.
        pass

    def thumbnail_status(self, previewpath):
        """Check the thumbnail status from the status file."""
        self._previewpath = previewpath
        
        if not os.access(self.filepath, os.F_OK):
            return "error"
        
        statusfile = self._outfile('status')
        if not os.access(statusfile, os.F_OK):
            return "completed"
        
        f = open(statusfile, "r")
        status = f.read().strip()
        f.close()
        return status

    def _outfile(self, suffix):
        return '%s.%s'%(self._previewpath, suffix)

    ##### Handler registration #####

    @classmethod
    def register(cls, names):
        def f(o):
            for name in names:
                if name in cls._handlers:
                    raise ValueError("""File handler %s already registered""" % name)
                cls._handlers[name] = o
            return o
        return f

    @classmethod
    def get_handler(cls, **kwargs):
        """Return an appropriate file handler."""
        handler = None

        filename = kwargs.get('filename')        
        if filename:
            # Ignore compression
            f = filename.split(".")
            if f[-1].lower() in ['gz', 'bz2', 'zip']:
                f.pop()
            # Use remaining file ext to find the handler
            if f:
                handler = f[-1].lower()
            
        handler = cls._handlers.get(handler, cls)
        return handler(**kwargs)
    
    
    
    
    
@BinaryHandler.register(['jpg', 'jpeg', 'png', 'gif', 'bmp', 'ai', 'pdf', 'eps', 'mpeg'])
class ImageHandler(BinaryHandler):
    def build_scale(self, img, outfile, tilesize=256):
        # ImageMagick...
        # convert -resize 128x128 -background white -gravity center -format jpg -quality 75 bdo:2010011400000  bdo:2010011400000.thumb.jpg
        args = []
        args.append('convert')
        
        args.append('-resize')
        args.append('%sx%s'%(tilesize, tilesize))

        args.append('-gravity')
        args.append('center')
        
        args.append('-format')
        args.append('jpg')
        
        args.append('-quality')
        args.append('80')
        
        if tilesize <= 128:
            args.append('-extent')
            args.append('%sx%s'%(tilesize, tilesize))

        args.append(img)
        args.append(outfile)

        a = subprocess.Popen(args)
        a.wait()
        

    def _thumbnail_build(self, workfile):
        self.build_scale(workfile, self._outfile('thumb.jpg'), tilesize=128)
        self.build_scale(workfile, self._outfile('small.jpg'), tilesize=512)
        self.build_scale(workfile, self._outfile('medium.jpg'), tilesize=1024)

    
        
        
if __name__ == "__main__":
    main(globals())
        
        
        



# $Id: user.py,v 1.114 2013/05/14 14:12:37 irees Exp $
"""User DBOs

Classes:
    BaseUser
    NewUser
    User
"""

import time
import operator
import hashlib
import random
import re
import weakref
import traceback

# EMEN2 imports
import emen2.db.exceptions
import emen2.db.dataobject

# DBO that contains a password and email address
class BaseUser(emen2.db.dataobject.BaseDBObject):

    attr_public = emen2.db.dataobject.BaseDBObject.attr_public | set(['email', 'password'])
    attr_protected = emen2.db.dataobject.BaseDBObject.attr_protected | set(['email', 'password'])
    attr_required = set(['email', 'password'])

    def init(self, d):
        super(BaseUser, self).init(d)

        # Required initialization params
        self.__dict__['email'] = None
        self.__dict__['password'] = None

        # Secret takes the format:
        # action type, args, ctime for when the token is set, and secret
        self.__dict__['secret'] = None

    def _set_password(self, key, value):
        # This will always fail unless you're an admin
        #    setpassword requires either a password or auth token as arguments.
        self.setpassword(None, value)
        return set(['password'])
    
    def _set_email(self, key, value):
        # This will always fail unless you're an admin --
        #    setemail requires either a password or auth token as arguments.
        self.setemail(value)
        return set(['email'])

    def isowner(self):
        return super(BaseUser, self).isowner() or self._ctx.username == self.name

    ##### Password methods #####

    def _hashpassword(self, password):
        password = unicode(password or '')
        if len(password) == 40:
            return password
        return hashlib.sha1(unicode(password)).hexdigest()

    def validate_password(self, password):
        # Only root is allowed to have no password. All user accounts must have a password.
        password = unicode(password or '')
        if len(password) < 6 and self.name != 'root' and len(password) != 40:
            self.error("Password too short; minimum 6 characters required")
        return password

    def checkpassword(self, password):
        try:
            if self._ctx.checkadmin():
                return True
        except AttributeError:
            pass

        # This needs to work even if there is no Context set.
        if self.get('disabled'):
            self.error(e=emen2.db.exceptions.DisabledUserError)

        # Check the hashed password against the stored hashed password.
        if self.password and self._hashpassword(password) == self.password:
            return True
            
        # Sleep for 2 seconds any time there is a wrong password!!
        time.sleep(2)
        self.error(e=emen2.db.exceptions.AuthenticationError)

    def setpassword(self, oldpassword, newpassword, secret=None):
        # Check that we know the existing password, or an authentication secret
        # checkpassword will sleep 2 seconds for each failed attempt
        # checkpassword will raise exception for failed attempt
        if not (
            self.isnew()
            or self._checksecret('resetpassword', None, secret) 
            or self.checkpassword(oldpassword)
            ):
            self.error(e=emen2.db.exceptions.AuthenticationError)

        # You must be logged in as this user (or admin) AND know the old password.
        newpassword = self._hashpassword(self.validate_password(newpassword))
        self._set('password', newpassword, True)
        self._delsecret()

    def resetpassword(self):
        """Reset the user password. This creates an internal 'secret' token that can be used to reset a password.
        The secret should never be accessible via public methods.
        """
        # The second argument should be the email address
        self._setsecret('resetpassword', None)

    #################################
    # email setting/validation
    #################################

    def setemail(self, email, password=None, secret=None):
        email = self.validate_email(email)
        # Check that we know the existing password, or an authentication secret
        # Note that admin users always return True for _checksecret
        # Note: the auth token is bound both to the method (setemail) and the
        #     specific requested email address.
        # Note: email can be changed before first commit.
        
        if self.isnew():
            self._set('email', email, self.isowner())
        elif self._checksecret('setemail', email, secret):
            self._set('email', email, self.isowner())
            self._delsecret()
        elif self.checkpassword(password):
            self._setsecret('setemail', email)
        else:
            self.error(e=emen2.db.exceptions.AuthenticationError)

        return self.email

    #################################
    # Secrets for account password resets
    #################################

    def _set_secret(self, key, value, **kwargs):
        # Complicated.. put/puts will strip out secret.
        # You have to get/put directly to get or set the secret.
        pass

    def _checksecret(self, action, args, secret):
        # I only want this to work on certain subclasses. See: User
        return False

    def _setsecret(self, action, args):
        # secret is set using __dict__ (like _ctx/_ptest) because it's a secret, and not a normal attribute.
        if not hasattr(self, 'secret'):
            self.__dict__['secret'] = None

        if self.secret:
            if action == self.secret[0] and args == self.secret[1]:
                return

        import emen2.db.database
        secret = emen2.db.database.getrandomid()
        self.__dict__['secret'] = (action, args, secret, time.time())

    def _delsecret(self):
        self.__dict__['secret'] = None

    #################################
    # validation methods
    #################################

    def validate_email(self, email):
        # if not re.match('(\S+@\S+)', email):
        # After a long discussion in #python, it is impossible to validate
        #     emails other than checking for '@'
        # Note: Forcing emails to be stored as lower case.
        email = unicode(email or '').strip().lower()
        if not email or '@' not in email:
            self.error("Invalid email format '%s'"%email)
        return email


signupinfo = set(["name_first", "name_middle", "name_last", "comments",    "institution",
    "department", "address_street", "address_city", "address_state", "address_zipcode",
    "country", "uri", "phone_voice", "phone_fax"])

class NewUser(BaseUser):
    # displayname, userrec, and groups are unset when committing, so they can skip validation.
    attr_public = BaseUser.attr_public | set(['signupinfo'])

    def init(self, d):
        super(NewUser, self).init(d)
        # New users store signup info in a dict,
        #     which is committed as a 'person' record when approved
        self.__dict__['signupinfo'] = {}

    # Setters
    def _set_signupinfo(self, key, value):
        self.setsignupinfo(value)
        return set(['signupinfo'])

    def setsignupinfo(self, update):
        self._set('signupinfo', update)
        self.validate()

    def validate(self):
        super(NewUser, self).validate()

        # Check signupinfo
        required = set(["name_first","name_last"])
        newsignup = {}

        for param, value in self.signupinfo.items():
            if not value:
                continue
                
            # These will be transferred to a Record
            try:
                value = self.validate_param(param, value)
            except ValueError:
                emen2.db.log.info("NewUser Validation: Couldn't validate new user signup field %s: %s"%(param, value))
                continue

            newsignup[param] = value

        for param in required:
            if not newsignup.get(param):
                raise ValueError, "Required param %s"%param

        if self.signupinfo.get('child'):
            newsignup['child'] = self.signupinfo['child']

        # print "Validated signup...."
        # print newsignup
        self.__dict__['signupinfo'] = newsignup

    # def setContext(...)
    #    Don't let non-admin users read this?

class User(BaseUser):
    """User record. This contains the basic metadata information for a single user account, including username, password, primary email address, active/disabled, timestamps, and link to more complete user profile. Group membership is stored in Group instances, and set here by db.user.get by checking an index. If available during db.user.get, a copy of the profile record and the user's "displayname" will also be set.

    @attr name Username for logging in, first character must be a letter, no spaces
    @attr password SHA1 hashed password
    @attr disabled True if user is disabled, unable to login
    @attr privacy Privacy level; 1 conceals personal information from anonymous users, 2 conceals personal information from all users
    @attr record Record ID containing additional profile information
    @attr email Semi-validated email address
    @property groups Set by database when accessed
    @property userrec Copy of profile record; set by database when accessed
    @property displayname User "display name"; set by database when accessed
    """

    # displayname, userrec, and groups are unset when committing, so they can skip validation.
    attr_public = BaseUser.attr_public | set(['privacy', 'disabled', 'displayname', 'userrec', 'groups', 'record'])

    # These get set during setContext and cleared before commit
    userrec = property(lambda s:s._userrec)
    displayname = property(lambda s:s._displayname)
    groups = property(lambda s:s._groups)

    def init(self, d):
        super(User, self).init(d)
        # Enabled/disabled, privacy level, and record ID pointer
        self.__dict__['disabled'] = False
        self.__dict__['privacy'] = 0
        self.__dict__['record'] = None

    #################################
    # Setters
    #################################

    # These are unvalidated parameters because they cleared when committing
    # They return an empty set because they don't really modify the User.

    def _set_groups(self, key, value):
        self._set('_groups', value, True)
        return set()

    def _set_displayname(self, key, value):
        self._set('_displayname', value, True)
        return set()

    def _set_userrec(self, key, value):
        self._set('_userrec', value, True)
        return set()

    # Users can set their own privacy level
    def _set_privacy(self, key, value):
        value = int(value)
        if value not in [0,1,2]:
            self.error("User privacy setting may be 0, 1, or 2.")
        return self._set(key, value, self.isowner())

    # Only admin can change enabled/disabled or record reference

    def _set_disabled(self, key, value):
        value = bool(value)
        if self.name == self._ctx.username and value:
            self.error("Cannot disable self!")
        return self._set(key, value, self._ctx.checkadmin())

    def _set_record(self, key, value):
        return self._set(key, value, self._ctx.checkadmin())

    #################################
    # "Secret" Authentication tokens
    #################################

    # I only have this method available on User, and not BaseUser.
    def _checksecret(self, action, args, secret):
        try:
            if self._ctx.checkadmin():
                return True
        except Exception, e:
            pass

        if not hasattr(self, 'secret'):
            self.__dict__['secret'] = None

        # This should check expiration time...
        if action and secret and getattr(self, 'secret', None):
            if action == self.secret[0] and args == self.secret[1] and secret == self.secret[2]:
                return True

        return False

    #################################
    # Enable/disable
    #################################
    # These will go through setattr -> setitem -> set_{param}

    def enable(self):
        self.disabled = False

    def disable(self):
        self.disabled = True

    def setprivacy(self, privacy):
        self.privacy = privacy

    #################################
    # Displayname and profile Record
    #################################

    def getdisplayname(self, lnf=False, record=None):
        """Get the user profile record from the current Context"""
        if self.record is None:
            return unicode(self.name)

        if not self._userrec:
            if not record:
                record = self._ctx.db.record.get(self.record) or {}
            self._set('_userrec', record, True)

        # self._set('_displayname', d, True)
        return self._formatusername(lnf=lnf)

    def _formatusername(self, lnf=False):
        if not self._userrec:
            return self.name

        nf = self._userrec.get('name_first')
        nm = self._userrec.get('name_middle')
        nl = self._userrec.get('name_last')

        #if u["name_first"] and u["name_middle"] and u["name_last"]:
        if nf and nm and nl:
            if lnf:
                uname = "%s, %s %s" % (nl, nf, nm)
            else:
                uname = "%s %s %s" % (nf, nm, nl)

        elif nf and nl:
            if lnf:
                uname = "%s, %s" % (nl, nf)
            else:
                uname = "%s %s" % (nf, nl)

        elif nl:
            uname = nl

        elif nf:
            uname = nf

        else:
            return self.name

        return uname

    #################################
    # Access methods
    #################################

    def setContext(self, ctx, hide=True):
        super(User, self).setContext(ctx)
        admin = self._ctx.checkreadadmin()
        ctxuser = self._ctx.username
        p = {}

        # secret is never made available except through direct btree.get().
        # This may cause the secret to be reset at some points...
        p['secret'] = None

        # Defaults for cached attributes..
        p['_userrec'] = {}
        p['_displayname'] = self.name
        p['_groups'] = set()

        # If the user has requested privacy, we return only basic info
        hide = self.privacy or self._ctx.username == 'anonymous'
        if admin or ctxuser == self.name:
            hide = False
        else:
            # This is a bug that needs to be fixed.
            p['password'] = None

        # You must access the User directly to get these attributes

        # Hide basic details from anonymous users
        if hide:
            p['email'] = None
            p['record'] = None

        self.__dict__.update(p)
        self.__dict__['_displayname'] = self.getdisplayname()




__version__ = "$Revision: 1.114 $".split(":")[1][:-1].strip()



# $Id: template.py,v 1.16 2012/12/14 05:24:29 irees Exp $
import itertools

from emen2.web.view import View
import emen2.db.config



@View.register
class TemplateRender(View):
    '''Renders a template given its path.  To be usable, the template must set public to true in the global namespace.

        <%! ... %> defines code in the global namespace in a template
        variables:
            public = True # allow public access
            headers = {} # just in case you want headers
        template defs:
            <%def name='mimetype()'></%def>: set the mime type
    '''

    @View.add_matcher(r'^/tmpl/(?P<template>.+)/$', name='main')
    @View.add_matcher(r'^/tmpl-%s/(?P<template>.+)/$'%emen2.__version__, name='main/version')
    def main(self, template='/simple', **kwargs):
        makot = emen2.db.config.templates.get_template(template)

        self.ctxt['inherit'] = False
        if (self.db and self.db._getctx().checkadmin()) or getattr(makot.module, 'public', False):
            self.template = template
            self.headers = getattr(makot.module, 'headers', {})
        else:
            self.ctxt['content'] = '<b>Error, private template</b>'



        

__version__ = "$Revision: 1.16 $".split(":")[1][:-1].strip()



# $Id: exceptions.py,v 1.19 2012/07/28 06:31:17 irees Exp $
"""Exceptions

Exceptions:
    SecurityError
    SessionError
    AuthenticationError
    DisabledUserError
    ValidationError
    ExistingKeyError
    TimeError
"""

# Security Errors
class SecurityError(Exception):
    "Security error."
    code = 401

class SessionError(SecurityError):
    "Session expired."

class AuthenticationError(SecurityError):
    "Invalid account name or password."

class DisabledUserError(SecurityError):
    "Disabled user."

# Validation Errors
class ValidationError(ValueError):
    """Validation error."""

class ExistingKeyError(ValueError):
    "This account name or email is already in use."

# Time out
class TimeError(Exception):
    """Operation timed out."""

__version__ = "$Revision: 1.19 $".split(":")[1][:-1].strip()



# $Id: context.py,v 1.40 2013/05/01 23:38:40 irees Exp $
"""Security and authentication

Classes:
    Context: DBO for storing a login
    AnonymousContext: Special Context for anonymous users
    SpecialRootContext: Special Context for root
    ContextDB: BTree subclass for storing Contexts

"""

import time
import operator
import hashlib
import random
import re
import weakref
import traceback

# EMEN2 imports
import emen2.db.database
import emen2.db.exceptions
import emen2.db.proxy
import emen2.db.config


# This is temporarily here. It may be moved.
class Cacher(object):
    """Help keep track of calculated values."""
    def __init__(self):
        self.cache = {}

    def reset_cache(self):
        self.cache = {}

    def store(self, key, result):
        self.cache[key] = result

    def check(self, key):
        # key = tuple(args)
        # print "cache check:", key
        if key in self.cache:
            # print '\tfound:', self.cache[key]
            return True, self.cache[key]
        # print '\tnot found'
        return False, None
        
    

# Contexts do not use BaseDBObject since they are completely internal to the DB
class Context(object):
    """Defines a database context (like a session). After a user is authenticated
    a Context is created, and used for subsequent access."""

    attr_user = set()

    def __init__(self, db=None, username=None, user=None, groups=None, grouplevels=None, host=None, maxidle=604800):
        self.db = None
        self.setdb(db)

        # Context UUID
        self.name = emen2.db.database.getrandomid()

        # Context user information
        self.user = user or {}
        self.username = username
        self.groups = groups or set()
        self.grouplevels = grouplevels or {}


        # Client IP
        self.host = host

        # Context time
        self.time = emen2.db.database.getctime()
        self.utcnow = emen2.db.database.utcnow()

        # Maximum idle time before context expires
        self.maxidle = maxidle
        
        # Used for caching items
        self.cache = Cacher()

    def __getstate__(self):
        # copy the dict since we change it
        odict = self.__dict__.copy()
        for i in ['db', 'user', 'groups', 'grouplevels']:
            odict.pop(i, None)
        return odict

    def setdb(self, db=None):
        """Associate a DB connection with the context."""
        if not db:
            return
        if not isinstance(db, emen2.db.proxy.DBProxy):
            db = emen2.db.proxy.DBProxy(db=db, ctx=self)
        self.db = db

    def refresh(self, grouplevels=None, host=None, db=None):
        if host != self.host:
            raise emen2.db.exceptions.SessionError, "Session expired"

        t = emen2.db.database.getctime()
        if t > (self.time + self.maxidle):
            raise emen2.db.exceptions.SessionError, "Session expired"

        self.time = t
        self.utcnow = emen2.db.database.utcnow()
        self.cache = Cacher()
        self.grouplevels = grouplevels or {}
        self.setdb(db=db)

        self.user = {}
        self.grouplevels["anon"] = 0
        self.grouplevels["authenticated"] = self.grouplevels.get('authenticated', 0)
        self.groups = set(self.grouplevels.keys())

    def checkadmin(self):
        return "admin" in self.groups

    def checkreadadmin(self):
        return set(["admin", "readadmin"]) & self.groups

    def checkcreate(self):
        return set(["admin", "create"]) & self.groups


class SpecialRootContext(Context):
    def refresh(self, user=None, grouplevels=None, host=None, username=None, db=None, txn=None):
        self.name = None
        self.setdb(db=db)
        self.username = username or u'root'
        self.time = emen2.db.database.getctime()
        self.utcnow = emen2.db.database.utcnow()
        self.cache = Cacher()        
        self.groups = set(["admin"])
        self.grouplevels = {"admin":3}


class AnonymousContext(Context):
    def refresh(self, user=None, grouplevels=None, host=None, db=None, txn=None):
        self.name = None
        self.setdb(db=db)
        self.username = 'anonymous'
        self.time = emen2.db.database.getctime()
        self.utcnow = emen2.db.database.utcnow()
        self.cache = Cacher()
        self.groups = set(["anon"])
        self.grouplevels = {"anon":0}

__version__ = "$Revision: 1.40 $".split(":")[1][:-1].strip()